#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Defines a legend for displaying components.

:copyright: 2015 Agile Geoscience
:license: Apache 2.0
"""
# from builtins import object
from io import StringIO
import csv
import warnings

import numpy as np
from matplotlib import pyplot as plt
from matplotlib import patches
from matplotlib.colors import rgb2hex

from .component import Component
from . import utils
from .defaults import LEGEND


class LegendError(Exception):
    """
    Generic error class.
    """
    pass


class Decor(object):
    """
    A single display style. A Decor describes how to display a given set
    of Component properties.

    In general, you will not usually use a Decor on its own. Instead, you
    will want to use a Legend, which is just a list of Decors, and leave
    the Decors to the Legend.

    Args:
      params (dict): The parameters you want in the Decor. There must be a
        Component to attach the decoration to, and at least 1 other attribute.
        It's completely up to you, but you probably want at least a colour
        (hex names like #AAA or #d3d3d3, or matplotlib's English-language
        names listed at http://ageo.co/modelrcolour are acceptable.

        The only other parameter the class recognizes for now is 'width',
        which is the width of the striplog element.

    Example:
      my_rock = Component({ ... })
      d = {'component': my_rock, 'colour': 'red'}
      my_decor = Decor(d)
    """
    def __init__(self, params):
        for k, v in params.items():
            k = k.lower().replace(' ', '_')
            try:
                v = v.lower()
            except AttributeError:
                v = v
            setattr(self, k, v)

        if not getattr(self, 'component', None):
            raise LegendError("You must provide a Component to decorate.")

        if len(self.__dict__) < 2:
            raise LegendError("You must provide at least one decoration.")

        # Make sure we have a width, even if it's None.
        w = getattr(self, 'width', None)
        if w:
            self.width = float(self.width)
        else:
            self.width = None

        # Deal with American spelling.
        a = getattr(self, 'color', None)
        if a:
            setattr(self, 'colour', a)
            delattr(self, 'color')

        # Make sure the colour is correctly formatted, and allow
        # the use of matplotlib's English-language colour names.
        c = getattr(self, 'colour', None)
        if c is not None:
            if type(c) in [list, tuple, np.ndarray]:
                try:
                    self.colour = rgb2hex(c)
                except TypeError:
                    raise LegendError("Colour not recognized: " + c)
            elif c[0] != '#':
                try:
                    self.colour = utils.name_to_hex(c)
                except KeyError:
                    raise LegendError("Colour not recognized: " + c)
            elif len(c) == 4:
                try:
                    self.colour = c[:2] + c[1] + 2*c[2] + 2*c[3]
                except TypeError:
                    raise LegendError("Colour not recognized: " + c)
            else:
                pass  # Leave the colour alone. Could assert here.
        else:
            self.colour = None

    def __repr__(self):
        s = str(self)
        return "Decor({0})".format(s)

    def __str__(self):
        s = []
        for key in self.__dict__:
            t = "{key}='{value}'"
            s.append(t.format(key=key, value=self.__dict__[key]))
        return ', '.join(s)

    def __add__(self, other):
        if isinstance(other, self.__class__):
            result = [self, other]
            return Legend(result)
        elif isinstance(other, Legend):
            result = [self] + other.__list
            return Legend(result)
        else:
            raise LegendError("You can only add legends or decors.")

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        # Weed out empty elements
        s = {k: v for k, v in self.__dict__.items() if v}
        o = {k: v for k, v in other.__dict__.items() if v}

        # Compare
        if s == o:
            return True
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    # If we define __eq__ we also need __hash__ otherwise the object
    # becomes unhashable. All this does is hash the frozenset of the
    # keys. (You can only hash immutables.)
    def __hash__(self):
        return hash(frozenset(self.__dict__.keys()))

    @classmethod
    def random(cls, component):
        """
        Returns a minimal Decor with a random colour.
        """
        colour = np.random.rand(3,)
        return cls({'colour': colour, 'component': component})

    @property
    def rgb(self):
        """
        Returns an RGB triple equivalent to the hex colour.
        """
        return utils.hex_to_rgb(self.colour)

    def plot(self):
        """
        Make a simple plot of the Decor.

        Args:
        widths (bool): Whether to use the widths in the plot.
        height (int): A scalar for the height, in inches.

        Returns:
        None. Instead the function creates a plot object as a side-effect.
        """

        u = 1  # A bit arbitrary; some sort of scale

        plt.figure(figsize=(1, 1))
        ax = plt.subplot(111)
        rect1 = patches.Rectangle((0, 0),
                                  u, u,
                                  color=self.colour)
        ax.add_patch(rect1)
        plt.text(1.2*u, 0.5*u,
                 self.component.summary(),
                 fontsize=max(u, 15),
                 verticalalignment='center',
                 horizontalalignment='left')
        plt.xlim([0, u])
        plt.ylim([0, u])
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        ax.invert_yaxis()

        plt.show()


class Legend(object):
    """
    A look-up table to assist in the conversion of Components to
    a plot colour.

    Args:
        list_of_Decors (list): The decors to collect into a legend. In
            general, you will want to leave legend building to the constructor
            class methods, `Legend.default()`, and `Legend.from_csv(string)`.
            We can add others over time, such as `from_xls` and so on.
    """

    def __init__(self, list_of_Decors):
        self.table = [d.__dict__ for d in list_of_Decors]
        self.__list = list_of_Decors
        self.__index = 0
        self._iter = iter(self.__list)  # Set up iterable.

    def __repr__(self):
        s = str(self)
        return "Legend({0})".format(s)

    def __str__(self):
        s = [str(d) for d in self.__list]
        return '\n'.join(s)

    def __getitem__(self, key):
        if type(key) is slice:
            i = key.indices(len(self.__list))
            result = [self.__list[n] for n in range(*i)]
            return Legend(result)
        elif type(key) is list:
            result = []
            for j in key:
                result.append(self.__list[j])
            return Legend(result)
        else:
            return self.__list[key]

    def __setitem__(self, key, value):
        self.__list[key] = value

    def __iter__(self):
        return self

    def __next__(self):
        try:
            result = self.__list[self.__index]
        except IndexError:
            self.__index = 0
            raise StopIteration
        self.__index += 1
        return result

    def next(self):
        """
        Retains Python 2 compatibility.
        """
        return self.__next__()

    def __len__(self):
        return len(self.__list)

    def __contains__(self, item):
        if isinstance(item, Decor):
            for d in self.__list:
                if item == d:
                    return True
        if isinstance(item, Component):
            for d in self.__list:
                if item == d.component:
                    return True
        return False

    def __add__(self, other):
        if isinstance(other, self.__class__):
            result = self.__list + other.__list
            return Legend(result)
        elif isinstance(other, Decor):
            result = self.__list + [other]
            return Legend(result)
        else:
            raise LegendError("You can only add legends or decors.")

    @classmethod
    def default(cls):
        """
        Generate a default legend. No arguments.

        Returns:
            Legend: The legend stored in `defaults.py`.
        """
        return cls.from_csv(LEGEND)

    @classmethod
    def random(cls, list_of_Components):
        """
        Generate a random legend for a given list of components.

        Returns:
            Legend: A legend with random colours.
        """
        list_of_Decors = [Decor.random(r) for r in list_of_Components]
        return cls(list_of_Decors)

    @classmethod
    def from_csv(cls, string):
        """
        Read CSV text and generate a Legend.

        In the first row, list the properties. Precede the properties of the
        component with 'comp ' or 'component '. For example:

        colour,  width, comp lithology, comp colour
        #FFFFFF, 0, ,
        #F7E9A6, 3, Sandstone, Grey
        #FF99CC, 2, Anhydrite,
        ... etc

        Note:
            To edit a legend, the easiest thing to do is probably this:

            - `legend.to_csv()`
            - Edit the legend, call it `new_legend`.
            - `legend = Legend.from_csv(new_legend)`
        """
        try:
            f = StringIO(string)  # Python 3
        except TypeError:
            f = StringIO(unicode(string))  # Python 2

        r = csv.DictReader(f, skipinitialspace=True)
        list_of_Decors = []
        for row in r:
            d, component = {}, {}
            for (k, v) in row.items():
                if k[:4].lower() == 'comp':
                    prop = ' '.join(k.split()[1:])
                    component[prop] = v.lower()
                else:
                    d[k] = v.lower()
            d['component'] = Component(component)
            list_of_Decors.append(Decor(d))

        return cls(list_of_Decors)

    def to_csv(self):
        """
        Renders a legend as a CSV string.

        No arguments.

        Returns:
            str: The legend as a CSV.
        """
        # We can't delegate this to Decor because we need to know the superset
        # of all Decor properties. There may be lots of blanks.
        header = []
        component_header = []
        for row in self:
            for j in row.__dict__.keys():
                header.append(j)
            for k in row.component.__dict__.keys():
                component_header.append(k)
        header = set(header)
        component_header = set(component_header)

        header = set(header)
        component_header = set(component_header)
        header.remove('component')
        header_row = ''
        if 'colour' in header:
            header_row += 'colour,'
            header.remove('colour')
            has_colour = True
        for item in header:
            header_row += item + ','
        for item in component_header:
            header_row += 'component ' + item + ','

        # Now we have a header row! Phew.
        # Next we'll go back over the legend and collect everything.
        result = header_row.strip(',') + '\n'
        for row in self:
            if has_colour:
                result += row.__dict__.get('colour', '') + ','
            for item in header:
                result += str(row.__dict__.get(item, '')) + ','
            for item in component_header:
                result += str(row.component.__dict__.get(item, '')) + ','
            result += '\n'

        return result

    @property
    def max_width(self):
        """
        The maximum width of all the Decors in the Legend. This is needed
        to scale a Legend or Striplog when plotting with widths turned on.
        """
        maximum = max([row.width for row in self.__list])
        return maximum or 0

    def get_colour(self, c, default='#eeeeee', match_only=None):
        """
        Get the display colour of a component.

        Args:
           c (component): The component to look up.
           default (str): The colour to return in the event of no match.
           match_only (list of str): The component attributes to include in the
               comparison. Default: All of them.

        Returns:
           str. The hex string of the matching Decor in the Legend.
        """
        if c:
            if match_only:
                # Filter the component only those attributes
                c = Component({k: getattr(c, k) for k in match_only})
            for decor in self.__list:
                if c == decor.component:
                    return decor.colour
        return default

    def get_width(self, c, default=0, match_only=None):
        """
        Get the display width of a component.

        Args:
           c (component): The component to look up.
           default (float): The width to return in the event of no match.
           match_only (list of str): The component attributes to include in the
               comparison. Default: All of them.

        Returns:
           float. The width of the matching Decor in the Legend.
        """
        if c:
            if match_only:
                # Filter the component only those attributes
                c = Component({k: getattr(c, k) for k in match_only})
            for decor in self.__list:
                if c == decor.component:
                    return decor.width
        return default

    def get_component(self, colour, tolerance=0, default=None):
        """
        Get the component corresponding to a display colour. This is for
        generating a Striplog object from a colour image of a striplog.

        Args:
           colour (str): The hex colour string to look up.
           tolerance (float): The colourspace distance within which to match.
           default (component or None): The component to return in the event
           of no match.

        Returns:
           component. The component best matching the provided colour.
        """
        if not (0 <= tolerance <= np.sqrt(195075)):
            raise LegendError('Tolerance must be between 0 and 441.67')

        for decor in self.__list:
            if colour.lower() == decor.colour:
                return decor.component

        # If we're here, we didn't find one yet.
        r1, g1, b1 = utils.hex_to_rgb(colour)

        # Start with a best match of black.
        best_match = '#000000'
        best_match_dist = np.sqrt(r1**2. + g1**2. + b1**2.)

        # Now compare to each colour in the legend.
        for decor in self.__list:
            r2, g2, b2 = decor.rgb
            distance = np.sqrt((r2-r1)**2. + (g2-g1)**2. + (b2-b1)**2.)
            if distance < best_match_dist:
                best_match = decor.component
                best_match_dist = distance
                best_match_colour = decor.colour

        if best_match_dist <= tolerance:
            return best_match
        else:
            with warnings.catch_warnings():
                warnings.simplefilter("always")
                w = "No match found for {0} ".format(colour.lower())
                w += "with tolerance of {0}. Best match is ".format(tolerance)
                w += "{0}, {1}".format(best_match.summary(), best_match_colour)
                w += ", d={0}".format(best_match_dist)
                warnings.warn(w)

            return default

    def plot(self):
        """
        Make a simple plot of the legend.

        Simply calls Decor.plot() on all of its members.

        TODO: Build a more attractive plot.
        """
        for d in self.__list:
            d.plot()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Reader for template strings.

Replaces Mac OS newlines with \n characters.
"""
import re


def get_template(name):
    """
    Still unsure about best way to do this, hence cruft.
    """
    text = re.sub(r'\r\n', r'\n', name)
    text = re.sub(r'\{([FISDE°].*?)\}', r'{{\1}}', text)
    return text

__las = """~Version
VERS .              3.0       :CWLS LOG ASCII STANDARD - VERSION 3.0
WRAP .               NO       :ONE LINE PER DEPTH STEP
DLM  .            COMMA       :DELIMITING CHARACTER
PROG .      {prog:>12s}      :LAS Program name and version
CREA . {date:16s}       :LAS Creation date {{YYYY/MM/DD hh:mm}}

~Well
#MNEM .UNIT  DATA             DESCRIPTION
#---- ------ --------------   -----------------------------
STRT .M      {start:8.3f}         :START DEPTH
STOP .M      {stop:8.3f}         :STOP DEPTH
STEP .M      {step:9.4f}        :STEP
NULL .       {null:9.4f}        :NULL VALUE

WELL .       {well:20s}  :WELL
FLD  .       UNDEFINED             :FIELD
CTRY .       CA                    :COUNTRY

PROV .       NOVA SCOTIA           :PROVINCE
UWI  .       {uwi:21s} :UNIQUE WELL ID
LIC  .       {lic:21s} :LICENSE NUMBER

~Parameter
#MNEM .UNIT  VALUE            DESCRIPTION
#---- ------ --------------   -----------------------------

#Required parameters
RUN  .        ONE             :RUN NUMBER
PDAT .        GL              :PERMANENT DATUM
APD  .M       {apd:8.3f}        :ABOVE PERM DATUM
DREF .        KB              :DEPTH REFERENCE
EREF .M       {eref:8.3f}        :ELEVATION OF DEPTH

#Remarks
R1   .                        :REMARK LINE 1
R2   .                        :REMARK LINE 2
R3   .                        :REMARK LINE 3

{section}

{curve}"""

__section = """~{name}_Parameter
{short} .   {source:16s} : {name} source          {S}
{short}D.   MD               : {name} depth reference {S}

~{name}_Definition
{short}T.M                   : {name} top depth       {F}
{short}B.M                   : {name} base depth      {F}
{short}D.                    : {name} description     {S}

~{name}_Data | {name}_Definition
{data}"""

__curve = """~Curve
#MNEM .UNIT  LOG CODES        DESCRIPTION
#---- ------ --------------   -----------------------------
DEPT .m        00 001 00 00               :Depth Index - Measured Depth
LITH .         00 000 00 00               :See remarks

~Ascii
{logs}"""

las = get_template(__las)
section = get_template(__section)
curve = get_template(__curve)



#!/usr/bin/env python
# -*- coding: utf 8 -*-
"""
A vocabulary for parsing lithologic or stratigraphic decriptions.

:copyright: 2015 Agile Geoscience
:license: Apache 2.0
"""
from io import open
import json
import warnings
import re
from itertools import islice

from . import defaults

SPECIAL = ['synonyms', 'parts_of_speech', 'abbreviations']


class LexiconError(Exception):
    """
    Generic error class.
    """
    pass


class Lexicon(object):
    """
    A Lexicon is a dictionary of 'types' and regex patterns.

    Most commonly you will just load the default one.

    Args:
        params (dict): The dictionary to use. For an example, refer to the
            default lexicon in ``defaults.py``.
    """

    def __init__(self, params):
        for k, v in params.items():
            k = re.sub(' ', '_', k)
            setattr(self, k, v)

        # Make sure the special attributes are set.
        # We probably don't really need to do this.
        for attr in SPECIAL:
            if not getattr(self, attr, None):
                setattr(self, attr, None)

    def __repr__(self):
        return str(self.__dict__)

    def __str__(self):
        keys = self.__dict__.keys()
        counts = [len(v) for k, v in self.__dict__.items() if v]
        s = "Lexicon("
        for i in zip(keys, counts):
            s += "'{0}': {1} items, ".format(*i)
        s += ")"
        return s

    @classmethod
    def default(cls):
        """
        Makes the default lexicon, as provided in ``defaults.py``.

        Returns:
            Lexicon: The default lexicon.
        """
        return cls(defaults.LEXICON)

    @classmethod
    def from_json_file(cls, json_file):
        """
        Load a lexicon from a JSON file.

        Args:
            json_file (str): The path to a JSON dump.
        """
        with open('lexicon.json', 'r') as fp:
            return cls(json.load(fp))

    def find_word_groups(self, text, category, proximity=2):
        """
        Given a string and a category, finds and combines words into
        groups based on their proximity.

        Args:
            text (str): Some text.
            tokens (list): A list of regex strings.

        Returns:
            list. The combined strings it found.

        Example:
            COLOURS = [r"red(?:dish)?", r"grey(?:ish)?", r"green(?:ish)?"]
            s = 'GREYISH-GREEN limestone with RED or GREY sandstone.'
            find_word_groups(s, COLOURS) --> ['greyish green', 'red', 'grey']
        """
        f = re.IGNORECASE
        words = getattr(self, category)
        regex = re.compile(r'(\b' + r'\b|\b'.join(words) + r'\b)', flags=f)
        candidates = regex.finditer(text)

        starts, ends = [], []
        groups = []

        for item in candidates:
            starts.append(item.span()[0])
            ends.append(item.span()[1])
            groups.append(item.group().lower())

        new_starts = []  # As a check only.
        new_groups = []  # This is what I want.

        skip = False
        for i, g in enumerate(groups):
            if skip:
                skip = False
                continue
            if (i < len(groups)-1) and (starts[i+1]-ends[i] <= proximity):
                if g[-1] == '-':
                    sep = ''  # Don't insert spaces after hyphens.
                else:
                    sep = ' '
                new_groups.append(g + sep + groups[i+1])
                new_starts.append(starts[i])
                skip = True
            else:
                if g not in new_groups:
                    new_groups.append(g)
                    new_starts.append(starts[i])
                skip = False

        return new_groups

    def find_synonym(self, word):
        """
        Given a string and a dict of synonyms, returns the 'preferred'
        word. Case insensitive.

        Args:
            word (str): A word.

        Returns:
            str: The preferred word, or the input word if not found.

        Example:
            >>> syn = {'snake': ['python', 'adder']}
            >>> find_synonym('adder', syn)
            'snake'
            >>> find_synonym('rattler', syn)
            'rattler'

        TODO:
            Make it handle case, returning the same case it received.
        """
        if word and self.synonyms:
            # Make the reverse look-up table.
            reverse_lookup = {}
            for k, v in self.synonyms.items():
                for i in v:
                    reverse_lookup[i.lower()] = k.lower()

            # Now check words against this table.
            if word.lower() in reverse_lookup:
                return reverse_lookup[word.lower()]

        return word

    def expand_abbreviations(self, text):
        """
        Parse a piece of text and replace any abbreviations with their full
        word equivalents. Uses the lexicon.abbreviations dictionary to find
        abbreviations.

        Args:
            text (str): The text to parse.

        Returns:
            str: The text with abbreviations replaced.
        """
        if not self.abbreviations:
            raise LexiconError("No abbreviations in lexicon.")

        def chunks(data, SIZE=25):
            """
            Regex only supports 100 groups for munging callbacks. So we have to
            chunk the abbreviation dicitonary.
            """
            it = iter(data)
            for i in range(0, len(data), SIZE):
                yield {k: data[k] for k in islice(it, SIZE)}

        def cb(g):
            """Regex callback"""
            return self.abbreviations.get(g.group(0)) or g.group(0)

        # Special cases.

        # TODO: We should handle these with a special set of
        # replacements that are made before the others.
        text = re.sub(r'w/', r'wi', text)

        # Main loop.
        for subdict in chunks(self.abbreviations):
            regex = r'(\b' + r'\b)|(\b'.join(subdict.keys()) + r'\b)'
            text = re.sub(regex, cb, text)

        return text

    def get_component(self, text, required=False, first_only=True):
        """
        Takes a piece of text representing a lithologic description for one
        component, e.g. "Red vf-f sandstone" and turns it into a dictionary
        of attributes.

        TODO:
            Generalize this so that we can use any types of word, as specified
            in the lexicon.
        """
        component = {}

        for i, (category, words) in enumerate(self.__dict__.items()):

            # There is probably a more elegant way to do this.
            if category in SPECIAL:
                # There are special entries in the lexicon.
                continue

            groups = self.find_word_groups(text, category)

            if groups and first_only:
                groups = groups[:1]
            elif groups:
                # groups = groups
                pass
            else:
                groups = [None]
                if required:
                    with warnings.catch_warnings():
                        warnings.simplefilter("always")
                        w = "No lithology in lexicon matching '{0}'"
                        warnings.warn(w.format(text))

            filtered = [self.find_synonym(i) for i in groups]
            if first_only:
                component[category] = filtered[0]
            else:
                component[category] = filtered

        return component

    @property
    def categories(self):
        """
        Lists the categories in the lexicon, except the
        optional categories.

        Returns:
            list: A list of strings of category names.
        """
        keys = [k for k in self.__dict__.keys() if k not in SPECIAL]
        return keys



#!/usr/bin/env python
# -*- coding: utf 8 -*-
"""
Helper functions for the striplog package.

"""

from matplotlib import colors


def hex_to_name(hexx):
    """
    Convert hex to a color name, using matplotlib's colour names.

    Args:
        hexx (str): A hexadecimal colour, starting with '#'.

    Returns:
        str: The name of the colour, or None if not found.
    """
    for n, h in colors.cnames.items():
        if h == hexx.upper():
            return n
    return None


def name_to_hex(name):
    """
    Convert a color name to hex, using matplotlib's colour names.

    Args:
        name (str): The name of a colour, e.g. 'red'.

    Returns:
        str: The hex code for the colour.
    """
    return colors.cnames[name.lower()]


def rgb_to_hex(rgb):
    """
    Utility function to convert (r,g,b) triples to hex.
    http://ageo.co/1CFxXpO

    Args:
      rgb (tuple): A tuple or list of RGB values in the
        range 0-255 (i.e. not 0 to 1).

    Returns:
      str: The hex code for the colour.
    """
    h = '#%02x%02x%02x' % tuple(rgb)
    return h.upper()


def hex_to_rgb(hexx):
    """
    Utility function to convert hex to (r,g,b) triples.
    http://ageo.co/1CFxXpO

    Args:
        hexx (str): A hexadecimal colour, starting with '#'.

    Returns:
        tuple: The equivalent RGB triple, in the range 0 to 255.
    """
    h = hexx.strip('#')
    l = len(h)

    return tuple(int(h[i:i+l//3], 16) for i in range(0, l, l//3))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Defines intervals and rock for holding lithologies.

:copyright: 2015 Agile Geoscience
:license: Apache 2.0
"""
import time
import re
import warnings

from . import las
from . import templates
from .striplog import Striplog


class WellError(Exception):
    """
    Generic error class.
    """
    pass


class Extra(dict):
    """
    Helper class for making dict-like things.
    """
    def __init__(self, **kwargs):
        dict.__init__(self, kwargs)
        self.__dict__ = self


class Well(las.LASReader):
    """
    Well contains everything about the well. It inherits from
    las.LASReader.

    For example, well will contain header fields, curves, and
    other data from the 'basic' LAS file.

    We will then read supplementary LAS files, and/or maybe CSV
    data, into one main other attribute of Well: the striplog.

    If we start using pandas dataframes, this is the place to
    do it.

    Args:
        f (str): The path to an LAS file.
        null_subs (float): Something to substitute for the declared
            null value, which is probably -999.25. Often it's convenient
            to use np.nan.
        unknown_as_other (bool): Whether you'd like to load unknown
            sections as plain text blocks. A hack to cope with LAS3 files
            without having to handle arbitrary sections.

    Note:
        This module is not very general. It was written to support a very
        specific workflow. If it seems to be useful for other things,
        we can come back and try to generalize it.

        Essentially, this entire thing needs to be replaced by Something
        that properly supports the LAS 3.0 format, both reading and writing.
    """
    def __init__(self, f=None,
                 lexicon=None,
                 null_subs=None,
                 unknown_as_other=True):

        # First generate the parent object if possible.
        if f:
            super(Well, self).__init__(f, null_subs, unknown_as_other)

        # Add an empty striplog dict-like for later.
        self.striplog = Extra()

        # If we got an OTHER section, let's try loading it as striplogs...
        other = getattr(self, 'other', None)
        if other:
            f = re.MULTILINE
            pattern = re.compile(r'^(\~)(?=\w+?_Parameter)', flags=f)
            chunks = [i for i in filter(None, pattern.split(self.other))]

            # This is gross but I can't see how else to get the tilde
            # back on the strings when I do the re.split().
            nchunks = [a + b for a, b in zip(chunks[::2], chunks[1::2])]

            for section in nchunks:
                name = re.search(r'^\~(\w+?)_', section, flags=f).group(1)
                striplog = Striplog.from_las3(section, lexicon=lexicon)
                self.add_striplog(striplog, name.lower())

    # __repr__, __str__, etc, will come from LASReader

    def add_las(self, f, null_subs=None, unknown_as_other=True):
        """
        Add data from a LAS to the well object. Returns nothing.

        Args:
            f (str): The path to an LAS file.
            null_subs (float): Something to substitute for the declared
                null value, which is probably -999.25. Often it's convenient
                to use np.nan.
            unknown_as_other (bool): Whether you'd like to load unknown
                sections as plain text blocks. A hack to cope with LAS3 files
                without having to handle arbitrary sections.
        """
        pass

    def add_striplog(self, striplog, name):
        """
        Add a striplog to the well object. Returns nothing.

        Args:
            striplog (Striplog): A striplog object.
            name (str): A name for the log, e.g. 'cuttings', or 'Smith 2012'
        """
        setattr(self.striplog, name, striplog)

    def striplogs_to_las3(self, use_descriptions=False):
        """
        Form the LAS3 string.

        Notes:
            - It's debatable if this should be a ``striplog`` function, but it
                contains well-level information so I'm putting it here.
            - I can't decide whether to handle file writing, or just leave it
                to the user. So all this does for now is return a string.
        """
        data = ''
        for name, striplog in self.striplog.items():
            if name[3].lower() in 'aeiou':
                short = re.sub(r'[aeiou]', '', name)[:4].upper()
            else:
                short = name[:4].upper()
            name = name.capitalize()
            this_data = striplog.to_csv(use_descriptions=use_descriptions,
                                        header=False)
            template = templates.section
            data += template.format(name=name,
                                    short=short,
                                    source=striplog.source,
                                    data=this_data) + '\n'

        eref, apd = -999.25, -999.25
        if self.parameters.DREF.data == 'KB':
            try:
                eref = float(self.parameters.EREF.data)
            except AttributeError:
                with warnings.catch_warnings():
                    warnings.simplefilter("always")
                    w = "There is no EREF."
                    warnings.warn(w)

        if self.parameters.PDAT.data == 'GL':
            try:
                apd = float(self.parameters.APD.data)
            except AttributeError:
                with warnings.catch_warnings():
                    warnings.simplefilter("always")
                    w = "There is no APD."
                    warnings.warn(w)

        time_now = time.strftime("%Y/%m/%d %H:%M", time.gmtime())
        template = templates.las
        result = template.format(prog='striplog.py',
                                      date=time_now,
                                      start=self.start,  # NB from logs
                                      stop=self.stop,    # NB from logs
                                      step=-999.25,  # striplogs have no step
                                      null=self.null,
                                      well=self.well.WELL.data,
                                      uwi=self.well.UWI.data,
                                      lic=self.well.LIC.data,
                                      apd=apd,
                                      eref=eref,
                                      section=data,
                                      curve='')
        return result.strip() + '\n'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
LAS File Reader

The main class defined here is LASReader, a class that reads a LAS file
and makes the data available as a Python object.
"""

# Original code
# Copyright (c) 2011, Warren Weckesser
#
# Slowly updating to LAS 3.0
# Copyright (c) 2015, Agile Geoscience
#
# Licensed under the terms of the ISC License
# http://www.isc.org/downloads/software-support-policy/isc-license/
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# Python 2 & 3 compatibility
from __future__ import print_function
#from builtins import str

# Carry on...
import re
import keyword

import numpy as np


def is_identifier(s):
    """
    Helper function for finding LAS identifier strings.

    Returns:
        bool: Whether the string is an identifier.
    """
    if s in keyword.kwlist:
        return False
    return re.match(r'^[a-z_][a-z0-9_]*$', s, re.I) is not None


class LASError(Exception):
    """
    Generic error class.
    """
    pass


class LASItem(object):
    """
    This class is just a namespace, holding the attributes 'name',
    'units', 'data', 'value', and 'descr'.  'value' is the numerical
    value of 'data', if it has a numerical value (specifically, if
    int() or float() don't raise an exception when given the value
    of the 'data' attribute).

    A class method, from_line(cls, line), is provided to parse
    a line from a LAS file and create a LASItem instance.
    """
    def __init__(self, name, units='', data='', descr=''):
        self.name = name
        self.units = units
        self.data = data
        self.value = self.__convert_to_value(data)
        self.descr = descr

    def __repr__(self):
        s = str(self)
        return "LASItem(%s)" % s

    def __str__(self):
        s = ("name='%s', units='%s', data='%s', descr='%s'" %
             (self.name, self.units, self.data, self.descr))
        return s

    def __convert_to_value(self, s):
        """
        Converts strings to numbers, if possible.
        """
        try:
            value = int(s)
        except ValueError:
            try:
                value = float(s)
            except ValueError:
                value = s
        return value

    @classmethod
    def from_line(cls, line):
        """
        Gets data from a line.
        """
        first, descr = line.rsplit(':', 1)
        descr = descr.strip()
        name, mid = first.split('.', 1)
        name = name.strip()
        if mid.startswith(' '):
            # No units
            units = ''
            data = mid
        else:
            units_data = mid.split(None, 1)
            if len(units_data) == 1:
                units = units_data[0]
                data = ''
            else:
                units, data = units_data
        return LASItem(name=name, units=units, data=data.strip(),
                       descr=descr.strip())


class LASSection(object):
    """
    Represents a "section" of a LAS file.

    A section is basically a collection of items, where each item has the
    attributes 'name', 'units', 'data' and 'descr'.

    Any item in the section whose name is a valid Python identifier is
    also attached to the object as an attribute.  For example, if `s` is a
    LASSection instance, and the corresponding section in the LAS file
    contained this line:

    FD   .K/M3               999.9999        :  Fluid Density

    then the item may be referred to as `s.FD` (in addition to the longer
    `s.items['FD']`).

    Attributes
    ----------
    items : dict
        The keys are the item names, and the values are LASItem instances.
    names : list
        List of item names, in the order they were read from the LAS file.

    """
    def __init__(self):
        # Note: In Python 2.7, 'items' could be an OrderedDict, and
        # then 'names' would not be necessary--one could use items.keys().
        self.items = dict()
        self.names = []

    def add_item(self, item):
        """
        Helper function to add attributes.
        """
        self.items[item.name] = item
        self.names.append(item.name)
        if is_identifier(item.name) and not hasattr(self, item.name):
            setattr(self, item.name, item)

    def display(self):
        """
        Helper function to show items.
        """
        for name in self.names:
            item = self.items[name]
            namestr = name
            if item.units != '':
                namestr = namestr + (" (%s)" % item.units)
            print("%-16s %-30s [%s]" % (namestr, "'" + item.data + "'",
                                        item.descr))


class LASReader(object):
    """
    The LASReader class holds data from a LAS file.

    This reader only handles LAS 2.0 files (as far as I know).

    Constructor
    -----------
    LASReader(f, null_subs=None, unknown_as_other=False)

    f : file object or string
        If f is a file object, it must be opened for reading.
        If f is a string, it must be the filename of a LAS file.
        In that case, the file will be opened and read.

    Attributes for LAS Sections
    ---------------------------
    version : LASSection instance
        This LASSection holds the items from the '~V' section.

    well : LASSection instance
        This LASSection holds the items from the '~W' section.

    curves : LASection instance
        This LASSection holds the items from the '~C' section.

    parameters : LASSection instance
        This LASSection holds the items from the '~P' section.

    other : str
        Holds the contents of the '~O' section as a single string.

    data : numpy 1D structured array
        The numerical data from the '~A' section.  The data type
        of the array is constructed from the items in the '~C'
        section.

    Other attributes
    ----------------
    data2d : numpy 2D array of floats
        The numerical data from the '~A' section, as a 2D array.
        This is a view of the same data as in the `data` attribute.

    wrap : bool
        True if the LAS file was wrapped. (More specifically, this
        attribute is True if the data field of the item with the
        name 'WRAP' in the '~V' section has the value 'YES'.)

    vers : str
        The LAS version. (More specifically, the value of the data
        field of the item with the name 'VERS' in the '~V' section).

    null : float or None
        The numerical value of the 'NULL' item in the '~W' section.
        The value will be None if the 'NULL' item was missing.

    null_subs : float or None
        The value given in the constructor, to be used as the
        replacement value of each occurrence of `null_value` in
        the log data.  The value will be None (and no substitution
        will be done) if the `null_subs` argument is not given to
        the constructor.

    start : float, or None
        Numerical value of the 'STRT' item from the '~W' section.
        The value will be None if 'STRT' was not given in the file.

    start_units : str
        Units of the 'STRT' item from the '~W' section.
        The value will be None if 'STRT' was not given in the file.

    stop : float
        Numerical value of the 'STOP' item from the '~W' section.
        The value will be None if 'STOP' was not given in the file.

    stop_units : str
        Units of the 'STOP' item from the '~W' section.
        The value will be None if 'STOP' was not given in the file.

    step : float
        Numerical value of the 'STEP' item from the '~W' section.
        The value will be None if 'STEP' was not given in the file.

    step_units : str
        Units of the 'STEP' item from the '~W' section.
        The value will be None if 'STEP' was not given in the file.

    """

    def __init__(self, f, null_subs=None, unknown_as_other=False):
        """
        f can be a filename (str) or a file object.

        If 'null_subs' is not None, its value replaces any values in the data
        that matches the NULL value specified in the Version section of the LAS
        file.
        """

        # This is a hack to allow loading random LAS3-style sections
        # Of course it would be better to do this properly...
        self.unknown_as_other = unknown_as_other

        self.dlm = None
        self.null = None
        self.null_subs = null_subs
        self.start = None
        self.start_units = None
        self.stop = None
        self.stop_units = None
        self.step = None
        self.step_units = None

        self.version = LASSection()
        self.well = LASSection()
        self.curves = LASSection()
        self.parameters = LASSection()
        self.other = ''
        self.data = None

        self.__read_las(f, unknown_as_other)

        if self.data is not None:
            c = len(self.curves.items)
            self.data2d = self.data.view(float).reshape(-1, c)
            if null_subs is not None:
                self.data2d[self.data2d == self.null] = null_subs

    def __read_wrapped_row(self, f, n):
        """
        Read a "row" of data from the Ascii section of a "wrapped" LAS file.

        `f` must be a file object opened for reading.
        `n` is the number of fields in the row.

        Returns the list of floats read from the file.
        """
        depth = float(f.readline().strip(r', \t'))
        values = [depth]
        while len(values) < n:
            new_vals = [float(s.strip(r', \t')) for s in f.readline().split()]
            values.extend(new_vals)
        return values

    def __read_wrapped_data(self, f, dt):
        """
        Reads a block of wrapped data by repeatedly calling
        __read_wrapped_row().
        """
        data = []
        ncols = len(dt.names)
        while True:
            try:
                row = self.__read_wrapped_row(f, ncols)
            except Exception:
                break
            data.append(tuple(row))
        data = np.array(data, dtype=dt)
        return data

    def __read_las(self, f, unknown_as_other):
        """
        Read an LAS file.

        Returns a dictionary with keys 'V', 'W', 'C', 'P', 'O' and 'A',
        corresponding to the sections of a LAS file.  The values associated
        with keys 'V', 'W', 'C' and 'P' will be lists of Item instances.  The
        value associated with the 'O' key is a string.

        The value associated with the 'A' key is a numpy structured array
        containing the log data.  The field names of the array are the
        mnemonics from the Curve section of the file.
        """

        # Allowed Column Data delimiters (DLM)s from Appendix II of
        # www.cwls.org/wp-content/uploads/2014/09/LAS_3_File_Structure.pdf
        delimiters = {'SPACE': ' ', 'TAB': '\t', 'COMMA': ','}

        opened_here = False

        try:
            if isinstance(f, basestring):
                opened_here = True
                f = open(f, 'r')
        except NameError:
            if isinstance(f, str):
                opened_here = True
                f = open(f, 'r')

        self.wrap = False

        line = f.readline()
        current_section = None
        current_section_label = ''
        while line and not line.startswith('~A'):   # Data.
            blank = re.search(r'\w', line) is None  # True if line is blank.
            comment = line.startswith('#')
            if not blank and not comment:
                # Then not blank or comment.
                if line.startswith('~'):  # Section definition.
                    if len(line) < 2:
                        raise LASError("Missing section character after '~'.")
                    current_section_label = line[1:2]
                    las3_flag = '_' in line
                    other = False
                    if current_section_label == 'V' and not las3_flag:
                        current_section = self.version
                    elif current_section_label == 'W' and not las3_flag:
                        current_section = self.well
                    elif current_section_label == 'C' and not las3_flag:
                        current_section = self.curves
                    elif current_section_label == 'P' and not las3_flag:
                        current_section = self.parameters
                    elif current_section_label == 'O' and not las3_flag:
                        current_section = self.other
                        other = True
                    elif unknown_as_other:
                        current_section = self.other
                        self.other += line
                        other = True
                    else:
                        raise LASError("Unknown section '%s'" % line)
                elif current_section is None:
                    raise LASError("Missing first section.")
                else:
                    if other:
                        # The 'Other' section is just lines of text, so we
                        # assemble them into a single string.
                        self.other += line
                        current_section = self.other
                    else:
                        # Parse the line into a LASItem and add it to the
                        # current section.
                        m = LASItem.from_line(line)
                        current_section.add_item(m)
                        # Check for the required items whose values we'll
                        # store as attributes of the LASReader instance.
                        if current_section == self.version:
                            if m.name == 'WRAP':
                                if m.data.strip() == 'YES':
                                    self.wrap = True
                            if m.name == 'VERS':
                                self.vers = m.data.strip()
                            if m.name.strip() == 'DLM':
                                self.dlm = delimiters[m.data.strip().upper()]
                        if current_section == self.well:
                            if m.name == 'NULL':
                                self.null = float(m.data)
                            elif m.name == 'STRT':
                                self.start = float(m.data)
                                self.start_units = m.units
                            elif m.name == 'STOP':
                                self.stop = float(m.data)
                                self.stop_units = m.units
                            elif m.name == 'STEP':
                                self.step = float(m.data)
                                self.step_units = m.units

            line = f.readline()

        if re.search(r'\w', line) is None:
            return

        # Finished reading the header--all that is left is the numerical
        # data that follows the '~A' line.  We'll construct a structured
        # data type, and, if the data is not wrapped, use numpy.loadtext
        # to read the data into an array.  For wrapped rows, we use the
        # function __read_wrapped() defined elsewhere in this module.
        # The data type is determined by the items from the '~Curves' section.
        dt = np.dtype([(name, float) for name in self.curves.names])
        if self.wrap:
            a = self.__read_wrapped_data(f, dt)
        else:
            if self.dlm:
                if self.dlm == ' ':
                    dlm = None
                else:
                    dlm = self.dlm
                a = np.loadtxt(f, delimiter=dlm, dtype=dt)
            else:
                a = np.loadtxt(f, dtype=dt)
        self.data = a

        if opened_here:
            f.close()


if __name__ == "__main__":
    import sys

    las = LASReader(sys.argv[1], null_subs=np.nan, unknown_as_other=False)
    print("wrap? ", las.wrap)
    print("vers? ", las.vers)
    print("dlm? ", las.dlm)
    print("null =", las.null)
    print("start =", las.start)
    print("stop  =", las.stop)
    print("step  =", las.step)
    print("Version ---")
    las.version.display()
    print("Well ---")
    las.well.display()
    print("Curves ---")
    las.curves.display()
    print("Parameters ---")
    las.parameters.display()
    print("Other ---")
    print(las.other)
    print("Data ---")
    print(las.data2d)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Defines components for holding properties of rocks or samples or whatevers.

:copyright: 2015 Agile Geoscience
:license: Apache 2.0
"""
import re


class ComponentError(Exception):
    """
    Generic error class.
    """
    pass


class Component(object):
    """
    Initialize with a dictionary of properties. You can use any
    properties you want e.g.:

        - lithology: a simple one-word rock type
        - colour, e.g. 'grey'
        - grainsize or range, e.g. 'vf-f'
        - modifier, e.g. 'rippled'
        - quantity, e.g. '35%', or 'stringers'
        - description, e.g. from cuttings

    You can include as many other things as you want, e.g.

        - porosity
        - cementation
        - lithology code
    """

    def __init__(self, properties):
        for k, v in properties.items():
            if k and v:
                setattr(self, k.lower(), v.lower())

    def __repr__(self):
        s = str(self)
        return "Component({0})".format(s)

    def __str__(self):
        s = []
        for key in self.__dict__:
            t = '"{key}":"{value}"'
            s.append(t.format(key=key, value=self.__dict__[key]))
        return ', '.join(s)

    def __bool__(self):
        if not self.__dict__.keys():
            return False
        else:
            return True

    # For Python 2
    __nonzero__ = __bool__

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        # Weed out empty elements
        s = {k: v for k, v in self.__dict__.items() if v}
        o = {k: v for k, v in other.__dict__.items() if v}

        # Compare
        if s == o:
            return True
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    # If we define __eq__ we also need __hash__ otherwise the object
    # becomes unhashable. All this does is hash the frozenset of the
    # keys. (You can only hash immutables.)
    def __hash__(self):
        return hash(frozenset(self.__dict__.keys()))

    @classmethod
    def from_text(cls, text, lexicon, required=None, first_only=True):
        """
        Generate a Component from a text string, using a Lexicon.

        Args:
            text (str): The text string to parse.
            lexicon (Lexicon): The dictionary to use for the
                categories and lexemes.
            first_only (bool): Whether to only take the first
                match of a lexeme against the text string.

        Returns:
            Component: A Component object, or None if there was no
                must-have field.
        """
        component = lexicon.get_component(text, first_only=first_only)
        if required and (required not in component):
            return None
        else:
            return cls(component)

    def summary(self, fmt=None, initial=True, default=''):
        """
        Given a format string, return a summary description of a component.

        Args:
            component (dict): A component dictionary.
            fmt (str): Describes the format with a string. Use '%'
                to signal a field in the component, which is analogous
                to a dictionary. If no format is given, you will
                just get a list of attributes.
            initial (bool): Whether to capitialize the first letter.
            default (str): What to give if there's no component defined.

        Returns:
            str: A summary string.

        Example:

            r = Component({'colour': 'Red',
                           'grainsize': 'VF-F',
                           'lithology': 'Sandstone'})

            r.summary()  -->  'Red, vf-f, sandstone'
        """
        if default and not self.__dict__:
            return default

        if not fmt:
            string, flist = '', []
            for item in self.__dict__:
                string += '{}, '
                flist.append(item)
            string = string.strip(', ')
        else:
            fmt = re.sub(r'%%', '_percent_', fmt)
            string = re.sub(r'\{(\w+)\}', '{}', fmt)
            string = re.sub(r'_percent_', '%', string)
            flist = re.findall(r'\{(\w+)\}', fmt)

        words = []
        for key in flist:
            word = self.__dict__.get(key.lower())
            if word and key[0].isupper():
                word = word.capitalize()
            if word and key.isupper():
                word = word.upper()
            if not word:
                word = ''
            words.append(word)

        try:
            summary = string.format(*words)
        except KeyError as e:
            raise ComponentError("No such attribute, "+str(e))

        if initial and summary:
            summary = summary[0].upper() + summary[1:]

        return summary



#!/usr/bin/env python
# -*- coding: utf 8 -*-
"""
==================
striplog
==================
"""
from .legend import Decor, Legend
from .interval import Interval
from .component import Component
from .well import Well
from .striplog import Striplog
from .lexicon import Lexicon

__all__ = ['Lexicon',
           'Component',
           'Decor',
           'Legend',
           'Interval',
           'Well',
           'Striplog']


__version__ = "unknown"
try:
    from ._version import __version__
except ImportError:
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Defines some default values for parsing cuttings descriptions.

:copyright: 2015 Agile Geoscience
:license: Apache 2.0
"""

LEGEND = """colour, width, comp lithology, comp colour, comp grainsize
#F7E9A6, 3, Sandstone, Grey, VF-F
#FF99CC, 2, Anhydrite, ,
#DBD6BC, 3, Heterolithic, Grey,
#FF4C4A, 2, Volcanic, ,
#86F0B6, 5, Conglomerate, ,
#FF96F6, 2, Halite, ,
#F2FF42, 4, Sandstone, Grey, F-M
#DBC9BC, 3, Heterolithic, Red,
#A68374, 2, Siltstone, Grey,
#A657FA, 3, Dolomite, ,
#FFD073, 4, Sandstone, Red, C-M
#A6D1FF, 3, Limestone, ,
#FFDBBA, 3, Sandstone, Red, VF-F
#FFE040, 4, Sandstone, Grey, C-M
#A1655A, 2, Siltstone, Red,
#363434, 1, Coal, ,
#664A4A, 1, Mudstone, Red,
#666666, 1, Mudstone, Grey, """

LEXICON = {'lithology': [r'overburden', r'sandstone', r'siltstone', r'shale',
                         r'mudstone', r'limestone', r'dolomite',
                         r'salt', r'halite', r'anhydrite', r'gypsum',
                         r'sylvite',
                         r'clay', r'mud', r'silt', r'sand', r'gravel',
                         r'boulders',
                         ],
           'amount': [r'streaks?', r'veins?', r'stringers?',
                      r'interbed(?:s|ded)?',
                      r'blotch(?:es)?', r'bands?', r'fragments?',
                      r'impurit(?:y|ies)',
                      r'minor', r'some', r'abundant', r'rare', r'flakes?',
                      r'[-\.\d]+%'
                      ],
           'grainsize': [r'vf(?:-)?', r'f(?:-)?', r'm(?:-)?', r'c(?:-)?',
                         r'vc',
                         r'very fine(?: to)?', r'fine(?: to)?',
                         r'medium(?: to)?', r'coarse(?: to)?', r'very coarse',
                         r'v fine(?: to)?', r'med(?: to)?', r'med.(?: to)?',
                         r'v coarse',
                         r'grains?', r'granules?', r'pebbles?', r'cobbles?',
                         r'boulders?',
                         ],
           'colour': [r"red(?:dish)?",
                      r"gray(?:ish)?",
                      r"grey(?:ish)?",
                      r"black(?:ish)?",
                      r"whit(?:e|ish)",
                      r"blu(?:e|ish)",
                      r"purpl(?:e|ish)",
                      r"yellow(?:ish)?",
                      r"green(?:ish)?",
                      r"brown(?:ish)?",
                      r"light", "dark",
                      r"sandy"
                      ],
           'synonyms': {'Overburden': ['Drift'],
                        'Anhydrite': ['Gypsum'],
                        'Salt': ['Halite', 'Sylvite'],
                        },
           'parts_of_speech': {'noun': ['lithology'],
                               'adjective': ['colour', 'grainsize'],
                               'subordinate': ['amount'],
                               },
           'abbreviations': {"gt": "gritty", "dist": "distillate", "gr": "grained", "LSD": "legal subdivision", "ptg": "parting", "alg": "algal", "mnr": "minor", "Assem": "Assem", "Vad": "vadose", "MCW": "mud cut water", "Pch": "patch", "gd": "good", "/": "with", "ga": "gauged", "bulb": "bulbous", "Var": "variation", "gn": "green", "alt": "altering", "Gil": "gilsonite", "Oyst": "oyster", "sch": "schist", "Clst": "claystone", "ves": "vesicular", "Nod": "nodules", "tt": "tightly", "V.P.S.": "very poor sample V.P.S.", "asph": "asphaltic", "SIP": "shut in pressure", "tn": "tan", "G&OCM": "gas and oil cut mud", "DI": "dual induction log", "Mol": "mollusca", "Inoc": "inoceramus*", "spec": "speckled", "@": "at", "Amph": "amphipora*", "vert": "vertical", "lrg": "larger", "md": "muddy", "wthrd": "weathered", "Brec": "breccia", "D & A": "dry and abandoned", "fau": "fauna", "depau": "depauperate", "shad": "shadow", "DF": "derrick floor", "gy": "gray", "Elev": "elevation", "indst": "indistinct", "Syring": "syringopora*", "fac": "faceted", "GL": "guard log", "brhg": "branching", "Gt": "grit", "Gr": "grains", "FIt": "fault", "Tas": "tasmanites*", "Spr": "spore", "fenst": "fenestral", "plag": "plagioclase", "p": "poorly", "Spk": "speck", "Spl": "sampole", "SNP": "sidewall neutron porosity log", "n.s.": "no sample", "Clcar": "calcarenite", "pbl": "pebble (4-64 mm)", "cmt": "cemented", "Spg": "sponge", "Grv": "gravel", "Cyp": "cypridopsis*", "Grt": "granite", "SP": "spontaneous potential", "Res": "residue", "Bioh": "bioherm", "Biost": "biostrom", "tex": "texture", "Brach": "brachiopod", "hydc": "hydrocarbon", "Ren": "renalcis*", "Biot": "biotite", "Peld": "pelletoid", "Bld": "boulder", "ter": "terriginous", "splty": "splintery", "acic": "acicular", "sug": "sugary", "bdd": "bedded", "bdg": "bedding", "gl": "glassy", "abd": "abundant", "Intst": "intersticies", "Micr": "micrite", "O str": "Ostracod", "Clcsp": "calcisphere", "bl": "bluish", "abs": "absent", "abt": "about", "conspic": "conspicuous", "Spfool": "superficial olite", "sli": "slightly", "med": "medium", "O&SW": "oil and salt water", "biocl": "bioclastic", "Agg": "aggregate", "DST": "drill stem test", "Mn": "manganese", "psdo": "pseudo", "len": "lentilcular", "rug": "rugosa", "Wd": "wood", "hkl": "hackly", "DLL": "dual laterolog", "t.s.": "thin section", "Mtrx": "matrix", "crpxln": "crystocrystalline", "Zn": "zone", "Anthr": "anthracite", "altg": "alternating", "blksh": "blackish", "bor": "boreding", "dru": "drusy", "ctd": "coated", "Conc": "concretion", "wtr cush": "water cushion", "ctc": "contact", "micropor": "microporosity", "Zr": "zircon", "Cono": "conodont", "sft": "soft", "Piso": "pisoid", "LL": "laterolog", "lmn": "limonitic", "volc": "volcanics", "Ctgs": "cuttings", "apr": "apparent", "tr": "trace", "app": "appear", "brt": "bright", "ls": "limestone", "Slt": "silt", "Circ": "circulate", "aph": "aphanitic", "oomol": "oomoldic", "brk": "brackish", "vit": "vitreous", "org": "organic", "n.v.p.": "no visible porosity", "w/": "with", "brd": "bored", "intxln": "intercrystalline", "Para": "paraparchites*", "Fp": "flowing pressure", "Ft": "foot", "perf": "perforated", "&": "and", "fen": "fenestraal", "pred": "predominantly", "Phlog": "phloaopite", "perm": "permeability", "fibr": "fibrous", "msm": "metasomatic", "cht": "chert", "Fe": "iron-ferruginous", "dns": "denser", "vi": "violet", "fros": "frosted", "calc": "calcitareous", "Fm": "formation", "pres": "preservation", "cntrt": "contorted", "wvy": "wavy", "vps": "very poor samples", "ooc": "oocastic", "musc": "muscovite", "glas": "glassy", "Sh": "shale", "lchd": "leached", "rhb": "rhombic", "glau": "glauconitic", "Brac": "brachiopod", "choc": "chocolate", "chit": "chitinous", "Str": "structure", "Sl": "slate", "clus": "cluster", "Sa": "salt", "Cmt": "cement", "lmpy": "lumpy", "Sd": "sand", "intrapar": "intraparticle", "phos": "phosphatic", "prly": "pearly", "Lstr": "lustre", "tns": "tension", "f": "finely", "fls": "flesh", "des": "descript", "SW": "salt water", "flt": "faulted", "Cbl": "cobble", "plty": "platy", "lithgr": "lithographic", "flk": "flake", "W.R.": "washed residue", "SO": "show of oil", "v": "very", "deb": "debris", "slty": "silty", "fld": "feldsparthic", "flg": "flaggy", "ptch": "patches", "Lig": "lignite", "mica": "micaeous", "bent": "bentonitic", "Min": "mineral", "sy-Ca": "sparry calcite", "Lim": "limonite", "Invtb": "invertebrate", "sps": "sparsly", "Mid": "middle", "yelsh": "ish", "sph": "spherules", "spl": "sample", "Ls": "limestone", "tab": "tabular", "Plcy": "palecypod", "scat": "scattered", "psool": "pseudo oolitic", "GR": "gamma ray", "spsly": "sparsly", "Bdeye": "birdseye", "purp": "purple", "Pyr": "pyrite", "hom": "homogeneous", "Kao": "kaolin", "Spic": "spicule", "hor": "horizontal", "Belm": "belemnites*", "sid": "sideritic", "aft": "after", "Rhb": "rhomb", "Typ": "type", "sim": "similar", "sil": "siliceous", "frag": "fragmental", "lam": "laminated", "fr": "fair", "mar": "maroon", "frac": "fractured", "srt": "sorting", "Dol": "dolomite", "I.P.": "in part", "cpct": "compact", "max": "maximum", "insl": "insoluble", "lac": "lacustrine", "Mrl": "marl", "mag": "magnetic", "Ctc": "contact", "ireg": "irregular", "lav": "lavender", "IAB": "initial air blow", "fl": "filled", "Tril": "trilobite", "Foram": "foraminifera", "MMCFG": "million cubic feet of gas", "Clvg": "cleavage", "Alg": "algal", "sp": "spotty", "Deer": "decrease", "Microspr": "microspar", "su": "sulphurous", "or": "orangish", "strk": "streaked", "stri": "striated", "sh": "shale", "Moll": "mollusc", "spkld": "speckled", "sm": "smooth", "sl": "slightly", "sc": "scales", "sb": "sub", "sa": "salt", "strg": "stringer", "Trip": "tripoli", "Equiv": "equivalent", "lse": "loose", "mnut": "minute", "BHCS": "bore hole compensated sonic", "Cal": "caliper", "Chk": "chalk", "Biomi": "biomicrite", "brit": "brittle", "coln": "colonial", "Smwt": "somewhat", "Rf": "reef", "Cht": "chert", "(D)": "development", "decr": "decreasing", "Rk": "rock", "RT": "rotary table", "fis": "fissile", "cln": "clean", "xl": "crystalline", "cbl": "cobble (64-256 mm)", "Aglm": "agglomerate", "Evap": "evapourite", "sa-c": "salt castic", "fib": "fibrous", "imp": "impression", "lt": "lighter", "eux": "euxinic", "cren": "crenulated", "gty": "gravity", "dess": "dessiccation", "clr": "clear", "TSTM": "too small to measure", "Bas": "basalt", "Shw": "show", "nod": "nodule", "uncons": "unconsolidated", "rexl": "recrystallization", "Cav": "cavernous", "FAB": "fair air blow", "aprox": "approximately", "\u20b5": "core", "Schm": "schist", "Shl": "shell", "xln": "crystalline", "Rad": "radial", "Oomol": "oomold", "Dist": "distillate", "extr": "extremely", "l": "lower", "Anhy": "anhydrite", "intv": "interval", "Stach": "stachyodes*", "crpxl": "cryptocrystalline", "Poln": "pollen", "Chtz": "chitinozoa", "MCFG": "thousand cubic feet of gas", "frmwk": "framework", "Lith": "lithology", "Macrofos": "macrofossil", "prphy": "porphyry", "Lut": "lutite", "OFM": "oil flecked mud", "Asph": "assemblage", "intpar": "interparticle", "Perm": "permeability", "Bdst": "boundstone", "chty": "cherty", "meta": "metamorphic", "ex": "excellent", "hrtl": "horizontal", "p-p": "pin point", "zeo": "zeolite", "n/s": "no show", "nac": "nacerous", "Wtr": "water", "apprx": "approximate", "rr": "rare", "Musc": "muscovite", "intclas": "intraclastic", "rep": "replacedment", "porcel": "porcelaneous", "BOPH": "barrels of oil per hour", "rd": "rounded", "rf": "reefoid", "sml": "small", "blk": "black", "IP": "initial production", "bld": "bladed", "Tub": "tube", "Pisol": "pisolite", "mtx": "matrix", "lmy": "limy", "Phos": "phosphate", "crd": "cored", "G": "gas", "Spo": "spore", "GCM": "gas cut mud", "SGCM": "slight gas cut mud", "OWWO": "old well worked over", "Novac": "novaculite", "rhmb": "rhombic", "argl": "argillate", "W": "west", "carb": "carbonaceous", "Frac": "fracture", "freq": "frequent", "ES": "electric", "Pyrxn": "pyroxene", "g": "good", "Contam": "contamination", "srtg": "Sorteding", "Chal": "chalcedony", "fspr": "feldsparathic", "Char": "charophyte", "SSO": "slight show of oil", "Hyde": "hydrocarbon", "w": "well", "intgn": "inter grown", "vrtb": "vertebrate", "Sphal": "sphalerite", "intpt": "interpretation", "cche": "caliche", "Frg": "fringe", "sacc": "saccharoidal", "onc": "oncolites", "Slick slick": "slickenside", "Diagn": "diagenesis", "zn": "zone", "Imp": "impression", "sblit": "sublithic", "Prod": "production", "Incl": "inclusion", "Frag": "fragment", "sphal": "sphalerite", "Iran": "granule", "dtrl": "detritalus", "wthd": "weathered", "WCM": "water cut mud", "GCW": "gas cut water", "Pet": "petroleum", "Microstyl": "microstylolite", "Volc": "volcanic", "BWPH": "barrels of water per hour", "thru": "throughout", "intcl": "intraclasts", "Biosp": "biosparite", "Girv": "girvanella*", "OWDD": "oil well drilled deeper", "consol": "consolidated", "Fus": "fusulinid", "sz": "size", "grysh": "greyish", "mrlst": "marlstone", "Strk": "streak", "SGCW": "slight gas cut water", "crnk": "crinkled", "Rbl": "rubble", "Fuc": "fucoid", "bdeye": "birdseye", "gyp": "gypsumiferous", "orng": "orange", "Endo": "endothyra*", "BHFP": "bottom hole flow pressure", "Xl": "crystal", "stn": "staining", "Qtz": "quartz", "csg": "casing", "Chlor": "chlorite", "r": "rare", "mky": "milky", "str": "streak", "Onc": "oncolite", "Btm": "bottom", "Slst": "siltstone", "dissem": "disseminated", "spr": "sparry", "Pap": "paper", "grnt": "granite", "Par": "particle", "ang": "angular", "Descr": "description", "intgwn": "intergrown", "Stylio": "styliolina*", "Intclas": "intraclast", "Rem": "remains", "chlor": "chlorite", "euhd": "euhedral", "Pend": "pendularous", "Rec": "recovery", "grnl": "granule (2-4 mm)", "Calc": "calcite", "p.d": "pressure deformation", "recem": "recemented", "strgr": "stringer", "lig": "lignitic", "Glauc": "glauconite", "intbd": "interbedded", "lim": "itic", "mic": "micro", "Db": "diabase", "Port por": "porosity", "Ech": "echinoid", "lit": "lithic", "Fluor": "fluoresceince", "vrvd": "varved", "Arag": "aragonite", "vgt": "varigated", "Clus": "cluster", "mot": "mottled", "surf": "surface", "pos": "possibility", "plas": "plastic", "pyr": "pyritized", "In": "inch", "kao": "kaolin", "Orbit": "Orbitolina", "MLL": "microlaterolog", "sdy": "sandy", "Mic": "micaceous", "Plt": "plant", "assoc": "associated", "rthy": "earthy", "Ivan": "ivanovia*", "suc": "sucrosic", "intercal": "intercalated", "glos": "glossy", "typ": "typical", "abv": "above", "MCO": "mud cut oil", "Unconf": "unconformity", "Bent": "bentonite", "dol": "dolomitic", "dom": "dominant", "flor": "fluorescence", "sltst": "siltstone", "brec": "brecciated", "Stri": "striae", "dism": "disseminated", "BHT": "bottom hole temperature", "exv": "extrusive", "Pol": "polish", "GTS": "gas to surface", "Mbr": "member", "exp": "exposed", "Pybit": "pyrobitumen", "a.a.": "same as above sample", "intlam": "interlaminated", "sd": "sand (1/16-2 mm)", "Rpl": "ripple", "dend": "dendritic", "pkr": "packer", "drlg": "drilling", "Spher": "spherule", "Hal": "halitiferous", "lstr": "lustre", "Exclas": "extraclast", "ML": "microlog, minilog", "anhy": "anhydritic", "drlr": "driller", "Rud": "rudist", "grapst": "grapestone", "Bubl": "bubble", "pris": "prismatic", "Bnd": "band", "och": "ochre", "spher": "spherulitic", "Jt": "joint", "occ": "occasional", "Wl": "well", "varic": "varicolored", "intrlam": "interlaminated", "O&G": "oil and gas", "mnrl": "mineralized", "Shlt por": "shelter porosity", "AOF": "absolute open flow", "Gyp": "gypsumiferous", "Bit": "bitumen", "Gast": "gastropod", "Pst": "pumice-stone", "PB": "plugged back", "Stromlt": "stromatolite", "BO": "barrels of oil", "dk": "darker", "CN": "compensated neutron", "KB": "kelly bushing", "dd": "dead", "Intvl": "interval", "stromlt": "stromatolite", "Fspr": "feldspar", "Milid": "miliolid", "cotg": "coateding", "cotd": "coateding", "repl": "replacement", "slily": "slightly", "struc": "structure", "SO&G": "show of oil and gas", "gen": "generally", "hetr": "heterogeneous", "crs": "coarse", "bar": "baritic", "bas": "basaltic", "(W)": "wildcat", "shy": "shaly", "FDL": "formation density log", "w/o": "without w/o", "BW": "barrels of water", "rsns": "resinous", "PL": "proximity log", "gran": "granular", "BWPD": "barrels of water per day", "x": "cross", "grad": "grading", "Qtzt": "quartzite", "crm": "cream", "res": "residuual", "Pt": "part", "OTD": "old total depth", "contam": "contaminated", "S.W.C.": "sidewall core", "qtz": "quartz", "magnt": "magnetite", "Pel": "pellet", "num": "numerous", "sec": "secondary", "fnly": "finly", "arg": "argillaceous", "ark": "arkosic", "rmn": "remainant", "prim": "primary", "volat": "volatile", "SOCW": "slight oil cut water", "piso": "pisolitic", "pkish": "pinkish", "metaph": "metamorphosed", "trnsp": "transparent", "irr": "irregular", "hornbd": "hornblend", "'' or do": "ditto", "biost": "biostromal", "gept": "geopetal", "trnsl": "translucent", "PD": "per day", "PH": "per hour", "SO&W": "show of oil and water", "Vnlet": "veinlet", "Tham": "thamnopora*", "wg": "vuggy", "C": "coal", "men": "meniscus", "exclas": "extraclastic", "lg": "long", "jt": "jointing", "comp": "completion", "gnsh": "greenish", "wk": "weak", "wi": "with", "wh": "white", "Exv": "extrusive rock", "S": "sonic, acoustilog", "magn": "magnetic", "gywk": "graywacke", "G.W.": "granite wash", "Satm sat": "saturation", "tub": "tubular", "tuf": "tuffaceous", "coq": "coquina", "vug": "vugular", "c": "coarsely", "fnt": "faintly", "dkr": "darker", "cov": "covered", "conch": "conchoidal", "intgran": "intergranular", "SI": "shut in", "Chit": "chitinous", "s": "small", "Meta": "metamorphic rock", "Sst": "sandstone", "brak": "brackish", "uni": "uniform", "com": "common", "cotd gn": "coated grains", "Het": "Heterostegina", "foram": "foraminiferal", "chky": "chalky", "cl": "clastic", "cb": "carbonized", "Glas": "glass", "rad": "radiating", "poly": "polygonal", "hvy": "heavy", "pol": "polished", "Pent": "pentamerus*", "Hem": "hematite", "tgh": "tough", "cp": "compare", "Surf": "surface", "ps": "pseudo-", "frg": "fringing", "pt": "partly", "Bur": "burrow", "fri": "friable", "blsh": "bluish", "pch": "patchy", "stmg": "streaming", "frs": "fresh", "skel": "skeletal", "pyrbit": "pyrobitumen", "pk": "pink", "devit": "devitrified", "authg": "authigenic", "pl": "plant", "cly": "clayey", "t.b.": "thin-bedded", "Tr": "trace", "IES": "induction electric", "WAB": "weak air blow", "lge": "large", "spic": "spicular", "psi": "pounds per square inch", "crbnt": "carbonate", "Tp": "top", "Pkst": "packstone", "Sedm": "sediment", "Cont": "content", "xbdg": "cross-bedding", "gns": "gneiss", "micgr": "micrograined", "Cl": "clay", "slt": "silt", "vcol": "varicolored", "undly": "underlying", "grnt.w": "granite wash", "n": "no, none, non", "contm": "contaminated", "sln": "solution", "rbl": "rubblbly", "fuc": "fucoidal", "Intr": "intrusive", "cvg": "cavings", "k": "permeabilityable", "slb": "slabby", "FTAB": "faint air blow", "DIL": "dual induction laterolog", "pyrcl": "pyroclastic", "cons": "considerable", "rndd": "rounded", "bot": "botryoidal", "Sel": "selenite", "vn": "vein", "Tf": "tuffaceous", "styl": "stylotitic", "conc": "concretionary", "mott": "mottled", "xlam": "cross-laminated", "x-strat": "cross-stratified", "Strom": "stromatoporoid", "ig": "igneous", "pap": "papery", "incr": "increasing", "litt": "littoral", "intstl": "interstitial", "bioturb": "bioturbated", "PPM": "parts per million", "GAP": "good air blow", "Repl": "replaced", "lith": "lithographic", "elong": "elongate", "Chara": "charophytes", "sat": "saturated", "incl": "inclusion", "Coq": "coquina", "Vug": "vug", "Cor": "coral", "intst": "intersticitial", "cncn": "concentric", "rng": "range", "orth": "orthoclase", "rdsh": "redish", "syn": "syntaxial", "Microfos": "microfossilferous", "phr": "phreatic", "Wkst": "wackestone", "pisol": "pisolitic", "Col": "color", "Jasp": "jasper", "Mat": "material", "Mbl": "marble", "intxl": "intercrystalline", "detr": "detrital", "sed": "sedimentary", "x-bd": "cross-bedded", "gsy": "grasy", "OWPB": "oil well plugged back", "min": "mineralized", "Sol": "Soution", "Vn": "vein", "col": "colored", "x-lam": "cross-laminated", "thn": "thin", "thk": "thick", "fltg": "floating", "Ig": "igneous rock", "imbd": "imbedded", "ck": "choke", "BHP": "bottom hole pressure", "yel": "yellow", "Orth": "orthoclase", "sptd": "spottedy", "spty": "spottedy", "Sphaer": "sphaerocodium*", "Pbl": "pebble", "intfrag": "interfragmental", "Scaph": "scaphopod", "resd": "residual", "Bd": "bed", "Fe-mag": "ferro-magnesian", "sks": "slickensided", "rexlzd": "recrystallized", "Bm": "basement", "Glob": "globigerina*", "elg": "elongate", "unident": "unidentifiable", "Fau": "fauna", "Gal": "galeolaria*", "cub": "cubic", "Fac": "facet", "Glos": "gloss", "Gab": "gabbro", "bnd": "banded", "Oo": "ooid", "Gns": "gneiss", "amb": "amber", "strat": "strataified", "amm": "ammonite", "vis": "visible", "mos": "mosaic", "por": "poroussity", "uncons.": "unconsolidated", "embd": "embedded", "Dia": "diameter", "rnd": "rounded", "sbang": "subangular", "cntr": "centered", "mol": "moldic", "Cvg": "caving", "bit": "bitumeninous", "Micropor": "micropore", "S.W.": "salt water", "amt": "amount", "mod": "moderate", "Crin": "crinoidal", "Lyr": "layer", "brn": "brown", "boudg": "boudinage", "OC": "oil cut", "Microol": "micro-oolite", "Ltl": "little", "V.op": "valve open", "amor": "amorphous", "Ark": "arkose", "clas": "clastic", "Psool": "pseudo oolite", "Strat": "strata", "Bdg": "bedding", "bri": "bright", "drsy": "drusy", "tstg": "testing", "Scol": "scolecodonts", "crpld": "crumpled", "SOCM": "slight oil cut mud", "rec": "recovered", "Fvst": "favosites*", "sbrndd": "sub rounded", "p.p.": "pin-poin", "HO": "heavy oil", "olv": "olive", "Mdst": "mudstone", "s & p": "salt and pepper", "venn": "vermillon", "u": "upper", "gvl": "gravel", "fos": "fossiliferous", "pet": "petroleumiferous", "Clclt": "calcilutite", "OSR": "oil source rock", "Len": "lens", "pel": "pellet", "circ": "circulation", "prom": "prominently", "fol": "foliated", "peld": "pelletoidal", "Tent": "tentaculites*", "prob": "probably", "bd": "bed", "flky": "flaky", "bf": "buff", "LL8": "laterolog-8", "cgl": "conglomerate", "Gwke": "graywacke", "slky": "silky", "Grap": "graptolite", "Bor": "bored", "bo": "bophaceous", "Bot": "botryoid", "phen": "phenocrysts", "posa": "possible", "bu": "buff", "sel": "selenite", "Gran": "granule", "r.f.p": "rounded frosted pitted", "Clslt": "calcisiltite", "ss": "sandstone", "Uc": "underclay", "oo": "ooidal", "GOR": "gas-to-oil ratio", "Clcrd": "calcirudite", "euhed": "euhedral", "Solen": "solenopora*", "Euryamph": "euryamphipora*", "Ptg": "parting", "od": "odor", "eqnt": "equant", "Ost": "ostracod", "o": "oil", "Ceph": "cephalopod", "irid": "iridescent", "ox": "oxidized", "Casph": "calcisphaera*", "ti": "tight", "Foss": "fossiliferous", "sept": "septate", "marn": "marine", "Scs": "scarce", "op": "open", "chk": "chalky", "ahd": "anhedral", "Fen": "fenestra", "qtzc": "quartzitic", "sol": "solitary", "corln": "coralline", "mas": "massive", "ferr": "ferruginous", "Cub": "cube", "intr": "intrusionive", "Chaet": "chaetetes*", "OCM": "oil cut mud", "qtzs": "quartzose", "qtzt": "quartzite", "GIP": "good initial puff", "loc": "location", "blky": "blocky", "phyl": "phyllitic", "vrtl": "vertical", "Ooc": "oolicast", "anhed": "anhedral", "aren": "arenaceous", "Ool": "oolite", "bioh": "biohermal", "diagn": "diagenesisetic", "Pelec": "pelecypod", "biot": "biotite", "abnt": "abundant", "var": "variable", "hem": "hematitic", "gil": "gilsonite", "calctc": "calcitic", "clyst": "claystone", "fl/": "flowing", "bur": "burrowed", "ea": "earthy", "micr": "micritic", "grdg": "grading", "Flk": "flake", "hi": "high", "s&p": "salt & pepper", "Flo": "flora", "Deb": "debris", "Splin": "splintery", "hd": "hard", "coqid": "coquinaoid", "Grst": "grainstone", "Phyl": "phyllite", "med.": "medium", "microxln": "microcrystalline", "Plag": "plagioclase", "mat": "material, matter", "bldr": "boulder", "up": "upper", "Tex": "texture", "mdy": "muddy", "WIP": "weak initial puff", "olvn": "olivine", "ab": "above", "mrl": "marly", "orsh": "orangish", "Allo": "allochem", "m": "medium", "deer": "decreasing", "chal": "chalcedony", "ovgth": "overgrowth", "ap": "appears", "Sid": "siderite", "ind": "indurated", "pit": "pitted", "trip": "tripolic", "DL": "density log", "Sil": "silica", "Lam": "laminations", "F.Q.G.": "frosted quartz  grains", "Cgl": "conglomerate", "hex": "hexagonal", "wxy": "waxy", "gry": "greyish", "Fe-st": "ironstone", "oom": "oomoldic", "grd": "graded", "BHSIP": "bottom hole shut in pressure", "Mag": "magnetite", "ool": "oolitic", "Lat": "laterite", "mass": "massive", "T.D.": "total depth", "Stn": "stain", "sqz": "squeezed", "E": "east", "BOPD": "barrels of oil per day", "aglm": "agglomerate", "evap": "evapourititic", "fluor": "fluoresceincent", "est": "estimated", "Su": "sulphur", "lent": "lenticular", "stal": "stalactitic", "N": "Neutron", "wtr": "water", "dolst": "dolostone", "bcm": "becoming", "OTS": "oil to surface", "SAB": "strong air blow", "Bry": "bryozoa", "O-Qtz": "orthoquartzite", "Styl": "stylolite", "crinal": "crinoidal", "Brk": "break", "mrly": "marly", "Av": "average", "xbd": "cross-bedded", "Radax": "radiaxial", "swbd": "swabbed"}
           }



#!/usr/bin/env python
# -*- coding: utf 8 -*-
"""
A striplog is a sequence of intervals.

:copyright: 2015 Agile Geoscience
:license: Apache 2.0
"""
import re
from io import StringIO
import csv
import operator

import numpy as np
from PIL import Image
from matplotlib import pyplot as plt
from matplotlib import patches
from matplotlib.ticker import MultipleLocator, FormatStrFormatter

from .interval import Interval
from .legend import Legend
from . import utils
from . import templates


class StriplogError(Exception):
    """
    Generic error class.
    """
    pass


class Striplog(object):
    """
    A Striplog is a sequence of intervals.

    We will build them from LAS files or CSVs.

    Args:
        list_of_Intervals (list): A list of Interval objects.
        source (str): A source for the data. Default None.
    """
    def __init__(self, list_of_Intervals, source=None, order='auto'):

        if not list_of_Intervals:
            m = "Cannot create an empty Striplog."
            raise StriplogError(m)

        if order.lower()[0] == 'a':  # Auto
            # Might as well be strict about it
            if all([iv.base > iv.top for iv in list_of_Intervals]):
                order = 'depth'
                self.order = 'depth'
            elif all([iv.base < iv.top for iv in list_of_Intervals]):
                self.order = 'elevation'
            else:
                m = "Could not determine order from tops and bases."
                raise StriplogError(m)

        # Could tidy this up with a base class and inheritance;
        # problem is cannot access self until initialized.
        if order.lower()[0] == 'd':
            self.order = 'depth'
            # Sanity check
            fail = any([iv.base < iv.top for iv in list_of_Intervals])
            if fail:
                m = "Depth order specified but base above top."
                raise StriplogError(m)
            # Order force
            list_of_Intervals.sort(key=operator.attrgetter('top'))
            self.start = list_of_Intervals[0].top
            self.stop = list_of_Intervals[-1].base

        else:
            self.order = 'elevation'
            fail = any([iv.base > iv.top for iv in list_of_Intervals])
            if fail:
                m = "Elevation order specified but base above top."
                raise StriplogError(m)
            # Order force
            r = True
            list_of_Intervals.sort(key=operator.attrgetter('top'), reverse=r)
            self.start = list_of_Intervals[-1].base
            self.stop = list_of_Intervals[0].top

        self.source = source

        self.__list = list_of_Intervals
        self.__index = 0  # Set up iterable.

    def __repr__(self):
        l = len(self.__list)
        details = "start={start}, stop={stop}".format(**self.__dict__)
        return "Striplog({0} Intervals, {1})".format(l, details)

    def __str__(self):
        s = [str(i) for i in self.__list]
        return '\n'.join(s)

    # Could use collections but doing this with raw magics.
    # Set up Striplog as an array-like iterable.
    def __getitem__(self, key):
        if type(key) is slice:
            i = key.indices(len(self.__list))
            result = [self.__list[n] for n in range(*i)]
            return Striplog(result)
        elif type(key) is list:
            result = []
            for j in key:
                result.append(self.__list[j])
            return Striplog(result)
        else:
            return self.__list[key]

    def __delitem__(self, key):
        if (type(key) is list) or (type(key) is tuple):
            # Have to compute what the indices *will* be as
            # the initial ones are deleted.
            indices = [x-i for i, x in enumerate(key)]
            for k in indices:
                del self.__list[k]
        else:
            del self.__list[key]
        self.__set_start_stop()

    def __len__(self):
        return len(self.__list)

    def __setitem__(self, key, value):
        if not key:
            return
        try:
            for i, j in enumerate(key):
                self.__list[j] = value[i]
        except TypeError:
            self.__list[key] = value
        except IndexError:
            raise StriplogError("There must be one Interval for each index.")

    def __iter__(self):
        return iter(self.__list)

    def __next__(self):
        """
        Supports iterable.

        """
        try:
            result = self.__list[self.__index]
        except IndexError:
            self.__index = 0
            raise StopIteration
        self.__index += 1
        return result

    def next(self):
        """
        Retains Python 2 compatibility.
        """
        return self.__next__()

    def __contains__(self, item):
        for r in self.__list:
            if item in r.components:
                return True
        return False

    def __reversed__(self):
        return Striplog(self.__list[::-1])

    def __add__(self, other):
        if isinstance(other, self.__class__):
            result = self.__list + other.__list
            return Striplog(result)
        elif isinstance(other, Interval):
            result = self.__list + [other]
            return Striplog(result)
        else:
            raise StriplogError("You can only add striplogs or intervals.")

    def __set_start_stop(self):
        """
        Reset the start and stop
        """
        if self.order == 'depth':
            self.start = self[0].top
            self.stop = self[-1].base
        else:
            self.start = self[-1].base
            self.stop = self[0].top

    def __sort(self):
        """
        Sorts into 'natural' order: top-down for depth-ordered
        striplogs; bottom-up for elevation-ordered.

        Note the a striplog sorts with the built-in `sorted()`
        by interval thickness, hence the need for this function.
        """
        if self.order == 'depth':
            self.__list.sort(key=operator.attrgetter('top'))
        else:
            self.__list.sort(key=operator.attrgetter('top'), reverse=True)
        return self

    @classmethod
    def __intervals_from_loglike(self, loglike, offset=2):
        """
        Take a log-like stream of numbers or strings,
        and return two arrays: one of the tops (changes), and one of the
        values from the stream.

        Args:
            loglike (array-like): The input stream of loglike data.
            offset (int): Offset (down) from top at which to get lithology,
            to be sure of getting 'clean' pixels.

        Returns:
            ndarray: Two arrays, tops and values.
        """
        loglike = np.array(loglike)
        all_edges = loglike[1:] == loglike[:-1]
        edges = all_edges[1:] & (all_edges[:-1] == 0)

        tops = np.where(edges)[0]
        tops = np.append(0, tops)

        values = loglike[tops + offset]

        return tops, values

    @classmethod
    def from_csv(cls, text,
                 lexicon=None,
                 source='CSV',
                 dlm=',',
                 points=False,
                 abbreviations=False,
                 complete=False):
        """
        Convert a CSV string into a striplog. Expects 2 or 3 fields:
            top, description
            OR
            top, base, description

        Args:
            text (str): The input text, given by ``well.other``.
            lexicon (Lexicon): A lexicon, required to extract components.
            source (str): A source. Default: 'CSV'.
            dlm (str): The delimiter, given by ``well.dlm``. Default: ','
            points (bool): Whether to treat as points or as intervals.
            abbreviations (bool): Whether to expand abbreviations in the
                description. Default: False.
            complete (bool): Whether to make 'blank' intervals, or just leave
                gaps. Default: False.

        Returns:
            Striplog: A ``striplog`` object.

        Example:
            # TOP       BOT        LITH
            312.34,   459.61,    Sandstone
            459.71,   589.61,    Limestone
            589.71,   827.50,    Green shale
            827.60,   1010.84,   Fine sandstone

        Todo:
            Automatic abbreviation detection.
        """

        text = re.sub(r'(\n+|\r\n|\r)', '\n', text.strip())

        as_strings = []
        try:
            f = StringIO(text)  # Python 3
        except TypeError:
            f = StringIO(unicode(text))  # Python 2
        reader = csv.reader(f, delimiter=dlm, skipinitialspace=True)
        for row in reader:
            as_strings.append(row)

        result = {'tops': [], 'bases': [], 'descrs': []}

        for i, row in enumerate(as_strings):
            if len(row) == 2:
                row = [row[0], None, row[1]]

            # TOP
            this_top = float(row[0])

            # BASE
            # Base is null: use next top if this isn't the end.
            if not row[1]:
                if i < len(as_strings)-1:
                    this_base = float(as_strings[i+1][0])  # Next top.
                else:
                    this_base = this_top + 1  # Default to 1 m thick at end.
            else:
                this_base = float(row[1])

            # DESCRIPTION
            this_descr = row[2].strip()

            # Deal with making intervals or points...
            if not points:
                # Insert intervals where needed.
                if complete and (i > 0) and (this_top != result['bases'][-1]):
                    result['tops'].append(result['bases'][-1])
                    result['bases'].append(this_top)
                    result['descrs'].append('')
            else:
                this_base = None

            # ASSIGN
            result['tops'].append(this_top)
            result['bases'].append(this_base)
            result['descrs'].append(this_descr)

        # Build the list.
        list_of_Intervals = []
        for i, t in enumerate(result['tops']):
            b = result['bases'][i]
            d = result['descrs'][i]
            interval = Interval(t, b, description=d,
                                lexicon=lexicon,
                                abbreviations=abbreviations)
            list_of_Intervals.append(interval)

        return cls(list_of_Intervals, source=source)

    @classmethod
    def from_img(cls, filename, start, stop, legend,
                 source="Image",
                 offset=10,
                 pixel_offset=2,
                 tolerance=0):
        """
        Read an image and generate Striplog.

        Args:
            filename (str): An image file, preferably high-res PNG.
            start (float or int): The depth at the top of the image.
            stop (float or int): The depth at the bottom of the image.
            legend (Legend): A legend to look up the components in.
            source (str): A source for the data. Default: 'Image'.
            offset (Number): The percentage of the way across the image from
                which to extract the pixel column. Default: 10.
            pixel_offset (int): The number of pixels to skip at the top of
                each change in colour. Default: 2.
            tolerance (float): The Euclidean distance between hex colours,
                which has a maximum (black to white) of 441.67 in base 10.
                Default: 0.

        Returns:
            Striplog: The ``striplog`` object.
        """
        im = np.array(Image.open(filename))
        col = im.shape[1]/(100./offset)
        rgb = im[:, col, :3]
        loglike = np.array([utils.rgb_to_hex(t) for t in rgb])

        # Get the pixels and colour values at 'tops' (i.e. changes).
        pixels, hexes = cls.__intervals_from_loglike(loglike,
                                                     offset=pixel_offset)

        # Scale pixel values to actual depths.
        length = float(loglike.size)
        tops = [start + (p/length) * (stop-start) for p in pixels]
        bases = tops[1:] + [stop]

        # Get the components corresponding to the colours.
        comps = [legend.get_component(h, tolerance=tolerance) for h in hexes]

        list_of_Intervals = []
        for i, t in enumerate(tops):
            interval = Interval(t, bases[i], components=[comps[i]])
            list_of_Intervals.append(interval)

        return cls(list_of_Intervals, source="Image")

    @classmethod
    def from_array(cls, a,
                   lexicon=None,
                   source="",
                   points=False,
                   abbreviations=False):
        """
        Turn an array-like into a Striplog. It should have the following
        format (where `base` is optional):

            [(top, base, description),
             (top, base, description),
             ...
             ]

        Args:
            a (array-like): A list of lists or of tuples, or an array.
            lexicon (Lexicon): A language dictionary to extract structured
                objects from the descriptions.
            source (str): The source of the data. Default: ''.
            points (bool): Whether to treat as point data. Default: False.

        Returns:
            Striplog: The ``striplog`` object.
         """
        csv_text = ''
        for interval in a:
            interval = [str(i) for i in interval]
            if (len(interval) < 2) or (len(interval) > 3):
                raise StriplogError('Elements must have 2 or 3 items')
            descr = interval[-1].strip('" ')
            interval[-1] = '"' + descr + '"'
            csv_text += ', '.join(interval) + '\n'

        return cls.from_csv(csv_text,
                            lexicon,
                            source=source,
                            points=points,
                            abbreviations=abbreviations)

    @classmethod
    def from_las3(cls, string, lexicon=None,
                  source="LAS",
                  dlm=',',
                  abbreviations=False):
        """
        Turn LAS3 'lithology' section into a Striplog.

        Args:
            string (str): A section from an LAS3 file.
            lexicon (Lexicon): The language for conversion to components.
            source (str): A source for the data.
            dlm (str): The delimiter.
            abbreviations (bool): Whether to expand abbreviations.

        Returns:
            Striplog: The ``striplog`` object.

        Note:
            Handles multiple 'Data' sections. It would be smarter for it
            to handle one at a time, and to deal with parsing the multiple
            sections in the Well object.

            Does not read an actual LAS file. Use the Well object for that.
        """
        f = re.DOTALL | re.IGNORECASE
        regex = r'\~\w+?_Data.+?\n(.+?)(?:\n\n+|\n*\~|\n*$)'
        pattern = re.compile(regex, flags=f)
        text = pattern.search(string).group(1)

        s = re.search(r'\.(.+?)\: ?.+?source', string)
        if s:
            source = s.group(1).strip()

        return cls.from_csv(text, lexicon,
                            source=source,
                            dlm=dlm,
                            abbreviations=abbreviations)

    def to_csv(self, use_descriptions=False, dlm=",", header=True):
        """
        Returns a CSV string built from the summaries of the Intervals.

        Args:
            use_descriptions (bool): Whether to use descriptions instead
                of summaries, if available.
            dlm (str): The delimiter.
            header (bool): Whether to form a header row.

        Returns:
            str: A string of comma-separated values.
        """
        data = ''

        if header:
            data += '{0:12s}{1:12s}'.format('Top', 'Base')
            data += '  {0:48s}'.format('Lithology')

        for i in self.__list:
            if use_descriptions and i.description:
                text = i.description
            elif i.primary:
                text = i.primary.summary()
            else:
                text = ''
            data += '{0:9.3f}'.format(i.top)
            data += '{0}{1:9.3f}'.format(dlm, i.base)
            data += '{0}  {1:48s}'.format(dlm, '"'+text+'"')
            data += '\n'

        return data

    def to_las3(self, use_descriptions=False, dlm=",", source="Striplog"):
        """
        Returns an LAS 3.0 section string.

        Args:
            use_descriptions (bool): Whether to use descriptions instead
                of summaries, if available.
            dlm (str): The delimiter.
            source (str): The sourse of the data.

        Returns:
            str: A string forming Lithology section of an LAS3 file.
        """
        data = self.to_csv(use_descriptions=use_descriptions,
                           dlm=dlm,
                           header=False)

        return templates.section.format(name='Lithology',
                                        short="LITH",
                                        source=source,
                                        data=data)

    def to_log(self, step=1.0, start=None, stop=None, legend=None):
        """
        Return a fully sampled log from a striplog. Useful for crossplotting
        with log data, for example.

        Args:
            step (float): The step size. Default: 1.0.
            start (float): The start depth of the new log. You will want to
                match the logs, so use the start depth from the LAS file.
                Default: The start of the striplog.
            stop (float): The stop depth of the new log. Use the stop depth
                of the LAS file. Default: The stop depth of the striplog.
            legend (Legend): If you want the codes to come from a legend,
                provide one. Otherwise the codes come from the log, using
                integers in the other they are encountered. If you use a
                legend, they are assigned in the order of the legend.

        Returns:
            ndarray: Two ndarrays in a tuple, (depth, logdata).
        """
        # Make the preparations.
        if not start:
            start = self.start

        if not stop:
            stop = self.stop

        pts = np.floor((stop - start)/step)
        stop = self.start + step * pts
        depth = np.linspace(start, stop, pts+1)
        result = np.zeros_like(depth)

        # Make a look-up table for the log values.
        if legend:
            table = {j.component: i+1 for i, j in enumerate(legend)}
        else:
            table = {j[0]: i+1 for i, j in enumerate(self.top)}

        # Assign the values from the lookup table.
        for i in self:
            top_index = np.ceil((i.top-start)/step)
            base_index = np.ceil((i.base-start)/step)+1
            key = table.get(i.primary) or 0
            result[top_index:base_index] = key

        return depth, result

    def plot_axis(self,
                  ax,
                  legend,
                  ladder=False,
                  default_width=1,
                  match_only=None):
        """
        Plotting, but only the Rectangles. You have to set up the figure.
        Returns a matplotlib axis object.

        Args:
            ax (axis): The matplotlib axis to plot into.
            legend (Legend): The Legend to use for colours, etc.
            ladder (bool): Whether to use widths or not. Default False.
            default_width (int): A width for the plot if not using widths.
                Default 1.
            match_only (list): A list of strings matching the attributes you
                want to compare when plotting.

        Returns:
            axis: The matplotlib axis.
        """
        for i in self.__list:
            origin = (0, i.top)
            colour = legend.get_colour(i.primary, match_only=match_only)
            thick = i.base - i.top
            d = default_width

            if ladder:
                w = legend.get_width(i.primary, match_only=match_only) or d
                w = d * w/legend.max_width
            else:
                w = d

            rect = patches.Rectangle(origin, w, thick, color=colour)
            ax.add_patch(rect)

        return ax

    def plot(self,
             legend=None,
             width=1,
             ladder=False,
             aspect=10,
             interval=(1, 10),
             match_only=None):
        """
        Hands-free plotting.

        Args:
            legend (Legend): The Legend to use for colours, etc.
            width (int): The width of the plot, in inches. Default 1.
            ladder (bool): Whether to use widths or not. Default False.
            aspect (int): The aspect ratio of the plot. Default 10.
            interval (int or tuple): The (minor,major) tick interval for depth.
                Only the major interval is labeled. Default (1,10).
            match_only (list): A list of strings matching the attributes you
                want to compare when plotting.

        Returns:
            None: The plot is a side-effect.
        """
        fig = plt.figure(figsize=(width, aspect*width))

        # And a series of Rectangle patches for the striplog.
        ax = fig.add_axes([0, 0, 1, 1])

        if not legend:
            # Build a random-coloured legend.
            comps = [i[0] for i in self.top if i[0]]
            legend = Legend.random(comps)

        self.plot_axis(ax=ax,
                       legend=legend,
                       ladder=ladder,
                       default_width=width,
                       match_only=match_only)

        ax.set_xlim([0, width])
        ax.set_ylim([self.stop, self.start])
        ax.set_xticks([])

        if type(interval) is int:
            interval = (1, interval)

        minorLocator = MultipleLocator(interval[0])
        ax.yaxis.set_minor_locator(minorLocator)

        majorLocator = MultipleLocator(interval[1])
        majorFormatter = FormatStrFormatter('%d')
        ax.yaxis.set_major_locator(majorLocator)
        ax.yaxis.set_major_formatter(majorFormatter)

        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['bottom'].set_visible(False)
        ax.yaxis.set_ticks_position('left')
        ax.get_yaxis().set_tick_params(which='both', direction='out')

        ax.patch.set_alpha(0)

        plt.show()

        return None

    def depth(self, d):
        """
        Get the interval at a particular depth.

        Args:
            d (Number): The depth to query.

        Returns:
            Interval: The interval at that depth, or None if
                the depth is outside the striplog's range.
        """
        for iv in self:
            if iv.top <= d <= iv.base:
                return iv
        return None

    def find(self, search_term):
        """
        Look for a regex expression in the descriptions of the striplog.
        If there's no description, it looks in the summaries.

        If you pass a Component, then it will search the components, not the
        descriptions or summaries.

        Case insensitive.

        Args:
            search_term (string or Component): The thing you want to search
                for. Strings are treated as regular expressions.
        Returns:
            Striplog: A striplog that contains only the 'hit' Intervals.
        """
        hits = []
        for i, iv in enumerate(self):
            try:
                search_text = iv.description or iv.primary.summary()
                pattern = re.compile(search_term, flags=re.IGNORECASE)
                if pattern.search(search_text):
                    hits.append(i)
            except TypeError:
                if search_term in iv.components:
                    hits.append(i)
        return self[hits]

    def find_gaps(self, index=False):
        """
        Finds gaps in a striplog.

        Args:
            index (bool): If True, returns indices of intervals with
            gaps after them.

        Returns:
            Striplog: A striplog of all the gaps. A sort of anti-striplog.

        TODO:
            Could do something similar to find overlaps.
        """
        hits = []
        intervals = []

        if self.order == 'depth':
            one, two = 'base', 'top'
        else:
            one, two = 'top', 'base'

        for i, iv in enumerate(self[:-1]):
            next_iv = self[i+1]
            if getattr(iv, one) < getattr(next_iv, two):
                hits.append(i)

                top = getattr(iv, one)
                base = getattr(next_iv, two)
                iv_gap = Interval(top, base)
                intervals.append(iv_gap)

        if index and hits:
            return hits
        elif intervals:
            return Striplog(intervals)
        else:
            return

    def prune(self, limit=None, n=None, percentile=None):
        """
        Remove intervals below a certain limit thickness.

        Args:
            limit (float): Anything thinner than this will be pruned.
            n (int): The n thinnest beds will be pruned.
            percentile (float): The thinnest specified percentile will be
                pruned.
        """
        if not (limit or n or percentile):
            m = "You must provide a limit or n or percentile for pruning."
            raise StriplogError(m)
        if limit:
            prune = [i for i, iv in enumerate(self) if iv.thickness < limit]
        if n:
            prune = self.thinnest(n=n, index=True)
        if percentile:
            n = np.floor(len(self)*percentile/100)
            prune = self.thinnest(n=n, index=True)

        del self[prune]  # In place delete
        return self

    def anneal(self):
        """
        Fill in empty intervals by growing from top and base.
        """
        gaps = self.find_gaps(index=True)

        if not gaps:
            return

        for gap in gaps:
            before = self[gap]
            after = self[gap+1]

            if self.order == 'depth':
                t = (after.top-before.base)/2
                before.base += t
                after.top -= t
            else:
                t = (after.base-before.top)/2
                before.top += t
                after.base -= t

        # These were in-place operations so we don't return anything
        return

    def thickest(self, n=1, index=False):
        """
        Returns the thickest interval(s) as a striplog.

        Args:
            n (int): The number of thickest intervals to return. Default: 1.
            index (bool): If True, only the indices of the intervals are
                returned. You can use this to index into the striplog.

        Returns:
            Striplog: A striplog of all the gaps. A sort of anti-striplog.
        """
        s = sorted(range(len(self)), key=lambda k: self[k])
        indices = s[-n:]
        if index:
            return indices
        else:
            return self[indices]

    def thinnest(self, n=1, index=False):
        """
        Returns the thinnest interval(s) as a striplog.

        TODO:
            If you ask for the thinnest bed and there's a tie, you will
            get the last in the ordered list.
        """
        s = sorted(range(len(self)), key=lambda k: self[k])
        indices = s[:n]
        if index:
            return indices
        else:
            return self[indices]

    @property
    def cum(self):
        """
        Returns the cumulative thickness of all filled intervals.

        It would be nice to use sum() for this (by defining __radd__),
        but I quite like the ability to add striplogs and get a striplog
        and I don't think we can have both, its too confusing.

        Not calling it sum, because that's a keyword.
        """
        total = 0.0
        for i in self:
            total += i.thickness
        return total

    @property
    def mean(self):
        """
        Returns the mean thickness of all filled intervals.
        """
        return self.cum / len(self)

    @property
    def top(self):
        """
        Summarize a Striplog with some statistics.
        """
        all_rx = set([iv.primary for iv in self])
        table = {r: 0 for r in all_rx}
        for iv in self:
            table[iv.primary] += iv.thickness

        return sorted(table.items(), key=operator.itemgetter(1), reverse=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Defines intervals for holding components.

:copyright: 2015 Agile Geoscience
:license: Apache 2.0
"""

import re
import warnings
import numbers
from functools import total_ordering

from .component import Component


class IntervalError(Exception):
    """
    Generic error class.
    """
    pass


@total_ordering
class Interval(object):
    """
    Used to represent a lithologic or stratigraphic interval, or single point,
    such as a sample location.

    Initialize with a top (and optional base) and a description and/or
    an ordered list of components.

    Args:
        top (float): Required top depth. Required.
        base (float): Base depth. Optional.
        lexicon (dict): A lexicon. See documentation. Optional unless you only
            provide descriptions, because it's needed to extract components.
        max_component (int): The number of components to extract. Default 1.
        abbreviations (bool): Whether to parse for abbreviations.

    """
    def __init__(self, top, base=None,
                 description='',
                 lexicon=None,
                 components=None,
                 max_component=1,
                 abbreviations=False):

        self.top = float(top)
        if base:
            self.base = float(base)
        else:
            self.base = self.top

        self.description = str(description)

        if components:
            self.components = list(components)
        else:
            self.components = None

        if self.description and (not self.components):
            if lexicon:
                comps = self.__parse_description(lexicon,
                                                 max_component=max_component,
                                                 abbreviations=abbreviations)
                self.components = comps
            else:
                with warnings.catch_warnings():
                    warnings.simplefilter("always")
                    w = "You must provide a lexicon to generate "
                    w += "components from descriptions."
                    warnings.warn(w)
                self.components = []

    def __repr__(self):
        s = str(self)
        return "Interval({0})".format(s)

    def __str__(self):
        s = "top: {top}, base: {base}, "
        s += "description: '{description}', "
        s += "components: {components}"
        return s.format(**self.__dict__)

    # Let's make addition combine the intervals into a new one, or
    # add thickness to the interval (if )
    def __add__(self, other):
        if isinstance(other, self.__class__):
            top = min(self.top, other.top)
            base = max(self.base, other.base)
            prop = 100 * max(self, other).thickness / (base-top)
            d1 = max(self, other).description.strip(' .,')
            d2 = min(self, other).description.strip(' .,')
            d = '{:.1f}% {} with {:.1f}% {}'.format(prop, d1, 100-prop, d2)
            c = max(self, other).components + min(self, other).components
            return Interval(top, base, description=d, components=c)
        elif isinstance(other, Component):
            top = self.top
            base = self.base
            d = self.description + ' with ' + other.summary()
            c = self.components + [other]
            return Interval(top, base, description=d, components=c)
        elif isinstance(other, numbers.Number):
            top = self.top
            base = self.base + other
            d = self.description
            c = self.components
            return Interval(top, base, description=d, components=c)
        else:
            raise IntervalError("You can only add intervals or thicknesses.")

    # Must supply __eq__ and one other rich comparison for
    # the total_ordering function to provide the others.
    # Not sure if this should be comparing thicknesses, or
    # depths, or...?
    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if self.thickness == other.thickness:
                return True
        return False

    def __lt__(self, other):
        if isinstance(other, self.__class__):
            if self.thickness < other.thickness:
                return True
        return False

    @property
    def primary(self):
        """
        Convenience function returning the first component.

        Returns:
            Component. The first one in the list of components.
        """
        if self.components:
            return self.components[0]
        else:
            return None

    @property
    def thickness(self):
        """
        Returns the thickness of the interval.

        Returns:
            Float: The thickness.
        """
        return abs(self.base - self.top)

    @property
    def kind(self):
        """
        The type of Interval: a 'point' (where base = top),
        or an 'interval', where thickness > 0.

        Returns:
            str: Either 'point' or 'interval'.
        """
        if self.thickness > 0:
            return 'point'
        return 'interval'

    def summary(self, fmt=None, initial=False):
        """
        Returns a summary of the interval.

        Args:
            fmt (str): A format string. Optional.
            initial (bool): Whether to capitalize the first letter.

        Returns:
            str: An English-language summary.

        TODO:
            Allow formatting of the entire string, not just the rock.
        """
        s = [r.summary(fmt=fmt, initial=initial) for r in self.components]
        summary = " with ".join(s)
        return "{0:.2f} m of {1}".format(self.thickness, summary)

    @staticmethod
    def __split_description(text):
        """
        Split a description into parts, each of which can be turned into
        a single component.
        """
        # Protect some special sequences.
        t = re.sub(r'(\d) ?in\. ', r'\1 inch ', text)  # Protect.
        t = re.sub(r'(\d) ?ft\. ', r'\1 feet ', t)  # Protect.

        # Transform all part delimiters to 'with'.
        t = re.sub(r'\;?\.? ?((under)? \d+%) (?=\w)', r' with \1 ', t)
        t = re.sub(r'\. ', r' with ', t)

        # Split.
        f = re.IGNORECASE
        pattern = re.compile(r'.(?:with|contain(?:s|ing)?)', flags=f)
        parts = filter(None, pattern.split(t))

        return [i.strip() for i in parts]

    def __parse_description(self, lexicon,
                            max_component=1,
                            abbreviations=False):
        """
        Turns a description into a lists of components. The items in the
        list are in the order they were found in the description, which is
        usually order of importance.
        """
        if abbreviations:
            text = lexicon.expand_abbreviations(self.description)
        else:
            text = self.description

        components = []
        for p, part in enumerate(self.__split_description(text)):
            if p == max_component:
                break
            components.append(Component.from_text(part, lexicon))

        return components



#!/usr/bin/env python
# -*- coding: utf 8 -*-
"""
Python installation file.
"""
from setuptools import setup
import re

verstr = 'unknown'
VERSIONFILE = "striplog/_version.py"
with open(VERSIONFILE, "r")as f:
    verstrline = f.read().strip()
    pattern = re.compile(r"__version__ = ['\"](.*)['\"]")
    mo = pattern.search(verstrline)
if mo:
    verstr = mo.group(1)
    print("Version "+verstr)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

REQUIREMENTS = ['numpy',
                'pillow',
                'matplotlib'
                ]

CLASSIFIERS = ['Development Status :: 4 - Beta',
               'Intended Audience :: Science/Research',
               'Natural Language :: English',
               'License :: OSI Approved :: Apache Software License',
               'Operating System :: OS Independent',
               'Programming Language :: Python',
               'Programming Language :: Python :: 2.7',
               'Programming Language :: Python :: 3.4',
               ]

setup(name='striplog',
      version=verstr,
      description='Tools for making and managing well data.',
      url='http://github.com/agile-geoscience/striplog',
      author='Agile Geoscience',
      author_email='hello@agilegeoscience.com',
      license='Apache 2',
      packages=['striplog'],
      tests_require=['pytest'],
      download_url='https://github.com/agile-geoscience/tarball/'+verstr,
      install_requires=REQUIREMENTS,
      classifiers=CLASSIFIERS,
      zip_safe=False,
      )
